# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import logging
import traceback
from functools import wraps, partial
from threading import Thread, Lock, RLock


def async(target):
    """Decorator. Make a function asynchronous.

    The decorated function is executed in a differed thread.
    """

    @wraps(target)
    def wrapped(*args, **kwargs):
        Thread(target=target, args=args, kwargs=kwargs, daemon=True).start()

    return wrapped


def async_in_pool(pool):
    """Decorator. Make a function asynchronous in a specified pool.

    The decorated function is executed in the specified threads-pool.

    .. Usage::

        class MyClass:
            __MyPool = ThreadPoolExecutor(10)

            @async_in_pool(__MyPool)
            def do_some_task(self):
                pass

    """

    def decorator(target):
        @wraps(target)
        def wrapped(*args, **kwargs):
            pool.submit(target, *args, **kwargs)

        return wrapped

    return decorator


def locked_function(target=None, *, lock=None, blocking=True, timeout=-1):
    """Decorator. Make a *function* "synchronized".

    Only one thread at time can access the decorated function.

    :param target: the function to decorate
    :param lock: the lock to be used (if not specified an RLock is created)
    :param blocking: if True the lock-acquirement is blocking
    :param timeout: timeout for the lock-acquirement
    """

    # If called with (keywords) arguments
    if target is None:
        return partial(locked_function, lock=lock, blocking=blocking,
                       timeout=timeout)

    if lock is None:
        target.__lock__ = RLock()
    else:
        target.__lock__ = lock

    @wraps(target)
    def locked(*args, **kwargs):
        try:
            if target.__lock__.acquire(blocking=blocking, timeout=timeout):
                return target(*args, **kwargs)
            else:
                return
        finally:
            try:
                target.__lock__.release()
            except RuntimeError:
                pass

    return locked


def locked_method(target=None, *, blocking=True, timeout=-1):
    """Decorator. Make a *method* synchronized.

    Only one thread at time can access the decorated method.

    :param target: the function to decorate
    :param blocking: if True the lock-acquirement is blocking
    :param timeout: timeout for the lock-acquirement
    """

    # If called with (keywords) arguments
    if target is None:
        return partial(locked_method, blocking=blocking, timeout=timeout)

    # generate a lock_name like "__method_name_lock__"
    lock_name = '__' + target.__name__ + '_lock__'
    target.__meta_lock__ = Lock()

    @wraps(target)
    def locked(self, *args, **kwargs):
        with target.__meta_lock__:
            lock = getattr(self, lock_name, None)

            # If the lock is not defined, then define it
            if lock is None:
                lock = RLock()
                setattr(self, lock_name, lock)

        try:
            if lock.acquire(blocking=blocking, timeout=timeout):
                return target(self, *args, **kwargs)
            else:
                return
        finally:
            try:
                lock.release()
            except RuntimeError:
                pass

    return locked


def suppress_exceptions(target=None, *, log=True):
    """Decorator. Suppress all the exception in the decorated function.

    :param log: If True (the default) exceptions are logged as warnings.
    """

    if target is None:
        return partial(suppress_exceptions, print_exc=log)

    @wraps(target)
    def wrapped(*args, **kwargs):
        try:
            return target(*args, **kwargs)
        except Exception:
            logging.warning('Exception suppressed:\n' + traceback.format_exc())

    return wrapped


def memoize(callable_):
    """Decorator. Caches a function's return value each time it is called.

    If called later with the same arguments, the cached value is returned
    (not reevaluated).

    .. Note::
        This works for any callable object.
        The arguments are cached (as strings) in object.cache.
    """
    cache = callable_.cache = {}

    @wraps(callable_)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = callable_(*args, **kwargs)
        return cache[key]

    return memoizer



# -*- coding: utf-8 -*-
#
# Python porting of qtractor fade functions
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>

"""
Functor arguments normalization:

t = (x - x0) / (x1 - x0)   where 'x' is the time  (x0 initial, x1 final)
a = y1 - y0                where 'y' is the value (y0 initial, y1 final)
b = y0
"""

from enum import Enum

from lisp.core.util import FunctionProxy


def fade_linear(t, a, b):
    """Linear fade."""
    return a * t + b


def fadein_quad(t, a, b):
    """Quadratic (t^2) fade in: accelerating from zero velocity."""
    return a * (t ** 2) + b


def fadeout_quad(t, a, b):
    """Quadratic (t^2) fade out: decelerating to zero velocity."""
    return a * (t * (2 - t)) + b


def fade_inout_quad(t, a, b):
    """Quadratic (t^2) fade in-out: acceleration until halfway,
    then deceleration.
    """
    t *= 2
    if t < 1.0:
        return 0.5 * a * (t ** 2) + b
    else:
        t -= 1
        return 0.5 * a * (1 - (t * (t - 2))) + b


def ntime(time, begin, duration):
    """Return normalized time."""
    return (time - begin) / (duration - begin)


class FadeInType(Enum):
    Linear = FunctionProxy(fade_linear)
    Quadratic = FunctionProxy(fadein_quad)
    Quadratic2 = FunctionProxy(fade_inout_quad)


class FadeOutType(Enum):
    Linear = FunctionProxy(fade_linear)
    Quadratic = FunctionProxy(fadeout_quad)
    Quadratic2 = FunctionProxy(fade_inout_quad)



"""Utility module for importing and checking gi.repository packages once"""

# "Solution" for https://bugzilla.gnome.org/show_bug.cgi?id=736260

import sys
sys.modules["gi.overrides.Gst"] = None
sys.modules["gi.overrides.GstPbutils"] = None

import gi
gi.require_version('Gst', '1.0')
gi.require_version('GstPbutils', '1.0')
gi.require_version('GstApp', '1.0')

# noinspection PyUnresolvedReferences
from gi.repository import Gst, GstPbutils, GObject, GstApp



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QT_TRANSLATE_NOOP

from lisp.backend.media_element import ElementType, MediaType
from lisp.modules.gst_backend.gi_repository import Gst
from lisp.modules.gst_backend.gst_element import GstMediaElement


class AutoSink(GstMediaElement):
    ElementType = ElementType.Output
    MediaType = MediaType.Audio
    Name = QT_TRANSLATE_NOOP('MediaElementName', 'System Out')

    def __init__(self, pipe):
        super().__init__()

        self.auto_sink = Gst.ElementFactory.make('autoaudiosink', 'sink')
        pipe.add(self.auto_sink)

    def sink(self):
        return self.auto_sink


# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
# Copyright 2016 Thomas Achtner <info@offtools.de>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import namedtuple

from ola.OlaClient import OLADNotRunningException, OlaClient

from lisp.application import Application
from lisp.core.clock import Clock
from lisp.core.configuration import config
from lisp.core.has_properties import Property
from lisp.core.plugin import Plugin
from lisp.core.signal import Connection
from lisp.core.util import time_tuple
from lisp.cues.cue import Cue
from lisp.cues.cue_time import CueTime
from lisp.cues.media_cue import MediaCue
from lisp.plugins.timecode.timecode_settings import TimecodeCueSettings, \
    TimecodeSettings
from lisp.ui import elogging
from lisp.ui.settings.app_settings import AppSettings
from lisp.ui.settings.cue_settings import CueSettingsRegistry
from lisp.ui.ui_utils import translate

TcFormatDef = namedtuple('TimecodeDef', ['format', 'millis'])
TcFormat = {
    'FILM': TcFormatDef(format=OlaClient.TIMECODE_FILM, millis=1000 / 24),
    'EBU': TcFormatDef(format=OlaClient.TIMECODE_EBU, millis=1000 / 25),
    'SMPTE': TcFormatDef(format=OlaClient.TIMECODE_SMPTE, millis=1000 / 30)
}


class HRCueTime(CueTime):
    _Clock = Clock(30)  # 1000 /30  = 33.3333 milliseconds


class OlaTimecode:
    def __init__(self):
        try:
            self.__client = OlaClient()
        except OLADNotRunningException:
            self.__client = None

        self.__cue = None
        self.__cue_time = None

        self.__track = 0
        self.__hres = config['Timecode'].getboolean('hres')
        self.__format = TcFormat[config['Timecode']['format']].format
        self.__millis = TcFormat[config['Timecode']['format']].millis
        self.__replace_hours = False
        self.__last_frame = -1

    @property
    def cue(self):
        return self.__cue

    def status(self):
        return bool(self.__client)

    def start_timecode(self, cue):
        """Start the timecode, using the given cue."""
        # Test and create client, if needed, return on fail
        if not self.__client:
            try:
                self.__client = OlaClient()
            except OLADNotRunningException:
                logging.debug('TIMECODE: Cannot track cue, OLA not running.')
                return

        # Load cue settings, if enabled, otherwise return
        if not cue.timecode['enabled']:
            return

        # Stop the currently "running" timecode
        self.stop_timecode()

        # Reload format settings
        self.__hres = config['Timecode'].getboolean('hres')
        self.__format = TcFormat[config['Timecode']['format']].format
        self.__millis = TcFormat[config['Timecode']['format']].millis

        # Setup new cue and options
        self.__cue = cue
        self.__cue_time = HRCueTime(cue) if self.__hres else CueTime(cue)
        self.__replace_hours = cue.timecode['replace_hours']
        self.__track = cue.timecode['track']

        # Start watching the new cue
        self.__cue_time.notify.connect(
            self.__send_timecode, Connection.QtQueued)

    def stop_timecode(self, rclient=False, rcue=False):
        """Stop the timecode

        :param rclient: Reset the client
        :param rcue: Reset the cues
        """
        if self.__cue_time is not None:
            self.__cue_time.notify.disconnect(self.__send_timecode)

        self.__last_frame = -1
        if rclient:
            self.__client = None
        if rcue:
            self.__cue = None
            self.__cue_time = None

    def __send_timecode(self, time):
        tt = time_tuple(time)
        frame = int(tt[3] / self.__millis)

        if self.__hres:
            if self.__last_frame == frame:
                return
            self.__last_frame = frame

        try:
            if not self.__replace_hours:
                track = tt[0]
            else:
                track = self.__track

            self.__client.SendTimeCode(self.__format,
                                       track, tt[1], tt[2], frame)
        except OLADNotRunningException:
            self.stop_timecode(rclient=True, rcue=True)
            elogging.error(
                translate('Timecode', 'Cannot send timecode.'),
                details=translate('Timecode', 'OLA has stopped.'))
        except Exception as e:
            self.stop_timecode(rclient=True, rcue=True)
            elogging.exception('Cannot send timecode.', e, dialog=False)


class Timecode(Plugin):
    Name = 'Timecode'

    def __init__(self):
        super().__init__()
        self.__client = OlaTimecode()
        self.__cues = set()

        # Register a new Cue property to store settings
        Cue.register_property('timecode', Property(default={}))

        # Register cue-settings-page
        CueSettingsRegistry().add_item(TimecodeCueSettings, MediaCue)
        # Register pref-settings-page
        AppSettings.register_settings_widget(TimecodeSettings)

        # Watch cue-model changes
        Application().cue_model.item_added.connect(self.__cue_added)
        Application().cue_model.item_removed.connect(self.__cue_removed)

    def init(self):
        if not config['Timecode'].getboolean('enabled'):
            logging.info('TIMECODE: disabled by application settings')
        elif not self.__client.status():
            logging.info('TIMECODE: disabled, OLA not running')

    def reset(self):
        self.__cues.clear()
        self.__client.stop_timecode(rclient=True, rcue=True)

    def __cue_changed(self, cue, property_name, value):
        if property_name == 'timecode':
            if value.get('enabled', False):
                if cue.id not in self.__cues:
                    self.__cues.add(cue.id)
                    cue.started.connect(self.__cue_started, Connection.QtQueued)
            else:
                self.__cue_removed(cue)

    def __cue_added(self, cue):
        cue.property_changed.connect(self.__cue_changed)
        self.__cue_changed(cue, 'timecode', cue.timecode)

    def __cue_removed(self, cue):
        try:
            self.__cues.remove(cue.id)
            if self.__client.cue is cue:
                self.__client.stop_timecode(rcue=True)

            cue.started.disconnect(self.__cue_started)
            cue.property_changed.disconnect(self.__cue_changed)
        except KeyError:
            pass

    def __cue_started(self, cue):
        if config['Timecode'].getboolean('enabled'):
            if cue is not self.__client.cue:
                self.__client.start_timecode(cue)
        elif cue is self.__client.cue:
            self.__client.stop_timecode(rcue=True)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, Qt, QDataStream, QIODevice, \
    QT_TRANSLATE_NOOP
from PyQt5.QtGui import QKeyEvent, QContextMenuEvent, QMouseEvent
from PyQt5.QtWidgets import QTreeWidget, QHeaderView, qApp

from lisp.core.signal import Connection
from lisp.cues.cue_factory import CueFactory
from lisp.layouts.list_layout.cue_list_item import CueListItem
from lisp.layouts.list_layout.listwidgets import CueStatusIcon, PreWaitWidget, \
    CueTimeWidget, NextActionIcon, PostWaitWidget


# TODO: here we should build a custom qt model/view
from lisp.ui.ui_utils import translate


class CueListView(QTreeWidget):

    key_event = pyqtSignal(QKeyEvent)
    context_event = pyqtSignal(QContextMenuEvent)
    select_cue_event = pyqtSignal(QMouseEvent)
    drop_move_event = QtCore.pyqtSignal(int, int)
    drop_copy_event = QtCore.pyqtSignal(int, int)

    HEADER_NAMES = ['', '#',
                    QT_TRANSLATE_NOOP('ListLayoutHeader', 'Cue'),
                    QT_TRANSLATE_NOOP('ListLayoutHeader', 'Pre wait'),
                    QT_TRANSLATE_NOOP('ListLayoutHeader', 'Action'),
                    QT_TRANSLATE_NOOP('ListLayoutHeader', 'Post wait'),
                    '']
    HEADER_WIDGETS = [CueStatusIcon, None, None, PreWaitWidget, CueTimeWidget,
                      PostWaitWidget, NextActionIcon]

    def __init__(self, cue_model, parent=None):
        """
        :type cue_model: lisp.layouts.list_layout.cue_list_model.CueListModel
        """
        super().__init__(parent)
        self._model = cue_model
        self._model.item_added.connect(self.__cue_added, Connection.QtQueued)
        self._model.item_moved.connect(self.__cue_moved, Connection.QtQueued)
        self._model.item_removed.connect(self.__cue_removed, Connection.QtQueued)
        self._model.model_reset.connect(self.__model_reset)
        self.__item_moving = False

        self.setHeaderLabels(
            [translate('ListLayoutHeader', h) for h in CueListView.HEADER_NAMES])
        self.header().setDragEnabled(False)
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(QHeaderView.Fixed)
        self.header().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(2, QHeaderView.Stretch)

        self.setColumnWidth(0, 40)
        self.setColumnWidth(len(CueListView.HEADER_NAMES) - 1, 18)
        self.setSelectionMode(self.SingleSelection)
        self.setDragDropMode(self.InternalMove)
        self.setAlternatingRowColors(True)
        self.setVerticalScrollMode(self.ScrollPerItem)

        self.setIndentation(0)

        self.currentItemChanged.connect(self.__current_changed)

        self.__guard = False
        self.verticalScrollBar().rangeChanged.connect(self.__update_range)

    def dropEvent(self, event):
        # Decode mimedata information about the drag&drop event, since only
        # internal movement are allowed we assume the data format is correct
        data = event.mimeData().data('application/x-qabstractitemmodeldatalist')
        stream = QDataStream(data, QIODevice.ReadOnly)

        # Get the starting-item row
        start_index = stream.readInt()
        new_index = self.indexAt(event.pos()).row()
        if not 0 <= new_index <= len(self._model):
            new_index = len(self._model)

        if qApp.keyboardModifiers() == Qt.ControlModifier:
            cue = self._model.item(start_index)
            new_cue = CueFactory.clone_cue(cue)

            self._model.insert(new_cue, new_index)
        else:
            self._model.move(start_index, new_index)

    def contextMenuEvent(self, event):
        if self.itemAt(event.pos()) is not None:
            self.context_event.emit(event)
        else:
            super().contextMenuEvent(event)

    def keyPressEvent(self, event):
        self.key_event.emit(event)

        if qApp.keyboardModifiers() == Qt.ControlModifier:
            # Prevent items to be deselected
            event.ignore()
        else:
            super().keyPressEvent(event)

    def mousePressEvent(self, event):
        if qApp.keyboardModifiers() == Qt.ControlModifier:
            # For cue selection
            self.select_cue_event.emit(event)
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        if qApp.keyboardModifiers() == Qt.ControlModifier:
            # Prevent items to be deselected
            if self.state() == self.DragSelectingState:
                item = self.itemAt(event.pos())
                self.setCurrentItem(item)

    def __current_changed(self, current_item, previous_item):
        self.scrollToItem(current_item)

    def __cue_added(self, cue):
        item = CueListItem(cue)
        item.setFlags(item.flags() & ~Qt.ItemIsDropEnabled)

        self.insertTopLevelItem(cue.index, item)
        self.__init_item(item, cue)

        # Select the added item and scroll to it
        self.setCurrentItem(item)
        # Ensure that the focus is set
        self.setFocus()

    def __cue_moved(self, start, end):
        item = self.takeTopLevelItem(start)

        self.insertTopLevelItem(end, item)
        self.setCurrentItem(item)
        self.__init_item(item, self._model.item(end))

    def __cue_removed(self, cue):
        self.takeTopLevelItem(cue.index)

        index = cue.index
        if index > 0:
            index -= 1
        self.setCurrentIndex(self.model().index(index, 0))

    def __model_reset(self):
        self.reset()
        self.clear()

    def __init_item(self, item, cue):
        item.name_column = CueListView.HEADER_NAMES.index('Cue')
        for index, widget in enumerate(CueListView.HEADER_WIDGETS):
            if widget is not None:
                self.setItemWidget(item, index, widget(cue))

        self.updateGeometries()

    def __update_range(self, min_, max_):
        if not self.__guard:
            self.__guard = True
            self.verticalScrollBar().setMaximum(max_ + 1)
            self.__guard = False



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGroupBox, QHBoxLayout, QComboBox, QLabel, \
    QVBoxLayout

from lisp.modules.gst_backend.elements.alsa_sink import AlsaSink
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class AlsaSinkSettings(SettingsPage):
    ELEMENT = AlsaSink
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.devices = self._discover_pcm_devices()
        self.devices['default'] = 'default'

        self.deviceGroup = QGroupBox(self)
        self.deviceGroup.setTitle(translate('AlsaSinkSettings', 'ALSA device'))
        self.deviceGroup.setGeometry(0, 0, self.width(), 100)
        self.deviceGroup.setLayout(QHBoxLayout())
        self.layout().addWidget(self.deviceGroup)

        self.device = QComboBox(self.deviceGroup)
        self.device.addItems(self.devices.keys())
        self.device.setCurrentText('default')
        self.device.setToolTip(
            translate('AlsaSinkSettings', 'ALSA devices, as defined in an '
                                          'asound configuration file'))
        self.deviceGroup.layout().addWidget(self.device)

        self.label = QLabel(translate('AlsaSinkSettings', 'ALSA device'),
                            self.deviceGroup)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.deviceGroup.layout().addWidget(self.label)

    def enable_check(self, enable):
        self.deviceGroup.setCheckable(enable)
        self.deviceGroup.setChecked(False)

    def load_settings(self, settings):
        device = settings.get('device', 'default')

        for name, dev_name in self.devices.items():
            if device == dev_name:
                self.device.setCurrentText(name)
                break

    def get_settings(self):
        if not (
            self.deviceGroup.isCheckable() and not self.deviceGroup.isChecked()):
            return {'device': self.devices[self.device.currentText()]}

        return {}

    def _discover_pcm_devices(self):
        devices = {}

        with open('/proc/asound/pcm', mode='r') as f:
            for dev in f.readlines():
                dev_name = dev[7:dev.find(':', 7)].strip()
                dev_code = 'hw:' + dev[:5].replace('-', ',')
                devices[dev_name] = dev_code

        return devices



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import os

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLineEdit, \
    QPushButton, QGridLayout, QLabel, QListWidget, QDialogButtonBox, \
    QFileDialog, QMessageBox

from lisp.ui.ui_utils import translate
from .session import Session


class UriChangerDialog(QDialog):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.resize(600, 300)
        self.setLayout(QVBoxLayout())

        # FILE SELECT
        self.sessionLayout = QHBoxLayout()
        self.layout().addLayout(self.sessionLayout)

        self.sessionFileEdit = QLineEdit(self)
        self.sessionFileEdit.editingFinished.connect(self.session_init)
        self.sessionLayout.addWidget(self.sessionFileEdit)

        self.sessionFileSelect = QPushButton(self)
        self.sessionFileSelect.clicked.connect(self.session_select)
        self.sessionLayout.addWidget(self.sessionFileSelect)

        self.sessionReload = QPushButton(self)
        self.sessionReload.setIcon(QIcon.fromTheme('view-refresh'))
        self.sessionReload.clicked.connect(self.session_init)
        self.sessionLayout.addWidget(self.sessionReload)

        # REPLACE
        self.replaceLayout = QGridLayout()
        self.layout().addLayout(self.replaceLayout)

        self.currentLabel = QLabel(self)
        self.currentLabel.setAlignment(Qt.AlignCenter)
        self.replaceLayout.addWidget(self.currentLabel, 0, 0)

        self.replaceLabel = QLabel(self)
        self.replaceLabel.setAlignment(Qt.AlignCenter)
        self.replaceLayout.addWidget(self.replaceLabel, 0, 1)

        self.prefixList = QListWidget(self)
        self.prefixList.currentTextChanged.connect(self.prefix_changed)
        self.replaceLayout.addWidget(self.prefixList, 1, 0, 2, 1)

        self.newPrefixLayout = QVBoxLayout()
        self.replaceLayout.addLayout(self.newPrefixLayout, 1, 1)

        self.prefixEdit = QLineEdit(self)
        self.newPrefixLayout.addWidget(self.prefixEdit)
        self.replaceLayout.setAlignment(self.prefixEdit, Qt.AlignTop)

        self.replaceApply = QPushButton(self)
        self.replaceApply.clicked.connect(self.session_replace)
        self.newPrefixLayout.addWidget(self.replaceApply)
        self.newPrefixLayout.setAlignment(self.replaceApply, Qt.AlignRight)

        self.buttons = QDialogButtonBox(self)
        self.buttons.setStandardButtons(QDialogButtonBox.Cancel)
        self.buttons.rejected.connect(self.reject)
        self.layout().addWidget(self.buttons)

        self.saveButton = self.buttons.addButton(QDialogButtonBox.Save)
        self.saveButton.clicked.connect(self.session_save)

        self.retranslateUi()

        self.session = None

    def retranslateUi(self):
        self.currentLabel.setText(translate('UriChanger', 'Current'))
        self.replaceLabel.setText(translate('UriChanger', 'Replace'))
        self.sessionFileSelect.setText(translate('UriChanger', 'Session file'))
        self.sessionReload.setToolTip(translate('UriChanger', 'Reload'))
        self.replaceApply.setText(translate('UriChanger', 'Replace'))

    def prefix_changed(self, prefix):
        if self.prefixEdit.text() == '':
            self.prefixEdit.setText(prefix)
        else:
            # Search if the prefixEdit text is one of the listed prefixes
            lower = 0
            upper = self.prefixList.count() - 1
            search = self.prefixEdit.text()

            while lower <= upper:
                middle = (lower + upper) // 2
                item = self.prefixList.item(middle).text()

                if item < search:
                    lower = middle + 1
                elif item > search:
                    upper = middle - 1
                else:
                    self.prefixEdit.setText(prefix)
                    break

    def session_select(self):
        file, _ = QFileDialog.getOpenFileName(self, filter='*.lsp',
                                              directory=os.getenv('HOME'))
        if file != '':
            self.sessionFileEdit.setText(file)
            self.session_init()

    def session_analyze(self):
        self.prefixList.clear()
        self.prefixList.addItems(self.session.analyze())
        self.prefixList.sortItems()

    def session_init(self):
        file = self.sessionFileEdit.text()

        if os.path.exists(file):
            self.session = Session(file)
            self.session_analyze()
        elif file.strip() != '':
            QMessageBox.critical(self, translate('UriChanger', 'Error'),
                                 translate('UriChanger',
                                           'Session file "{}" not found'.format(
                                               file)))

    def session_replace(self):
        self.session.replace(self.prefixList.currentItem().text(),
                             self.prefixEdit.text())
        self.session_analyze()

    def session_save(self):
        file, ok = QFileDialog.getSaveFileName(self, filter='*.lsp',
                                               directory=os.getenv('HOME'))
        if ok:
            self.session.save(file)
            self.accept()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import os

from lisp.core.loading import load_classes
from lisp.ui import elogging

__MODULES = {}


def load_modules():
    for name, module in load_classes(__package__, os.path.dirname(__file__)):
        try:
            __MODULES[name] = module()
            elogging.debug('MODULES: Loaded "{0}"'.format(name))
        except Exception as e:
            elogging.exception('Failed "{0}" lading'.format(name), e)


def translations():
    base_path = os.path.dirname(os.path.realpath(__file__))
    for module in next(os.walk(base_path))[1]:
        i18n_dir = os.path.join(base_path, module, 'i18n')
        if os.path.exists(i18n_dir):
            yield os.path.join(i18n_dir, module)


def terminate_modules():
    for module_name in __MODULES:
        try:
            __MODULES[module_name].terminate()
            elogging.debug('MODULES: Terminated "{0}"'.format(module_name))
        except Exception as e:
            elogging.exception('Failed "{0}" termination'.format(module_name),
                               e)


def check_module(modname):
    return modname.lower() in [mod.lower() for mod in __MODULES]



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QLabel, QProgressBar

from lisp.core.signal import Connection
from lisp.core.util import strtime
from lisp.cues.cue import CueNextAction, CueState
from lisp.cues.cue_time import CueTime, CueWaitTime
from lisp.ui.ui_utils import pixmap_from_icon


class CueStatusIcon(QLabel):
    STYLESHEET = 'background: transparent; padding-left: 20px;'
    SIZE = 16

    def __init__(self, cue, *args):
        super().__init__(*args)
        self.setStyleSheet(CueStatusIcon.STYLESHEET)

        self.cue = cue
        self.cue.interrupted.connect(self._stop, Connection.QtQueued)
        self.cue.started.connect(self._start, Connection.QtQueued)
        self.cue.stopped.connect(self._stop, Connection.QtQueued)
        self.cue.paused.connect(self._pause, Connection.QtQueued)
        self.cue.error.connect(self._error, Connection.QtQueued)
        self.cue.end.connect(self._stop, Connection.QtQueued)

    def _start(self):
        self.setPixmap(pixmap_from_icon('led-running', self.SIZE))

    def _pause(self):
        self.setPixmap(pixmap_from_icon('led-pause', self.SIZE))

    def _error(self, *args):
        self.setPixmap(pixmap_from_icon('led-error', self.SIZE))

    def _stop(self):
        self.setPixmap(pixmap_from_icon('', self.SIZE))

    def sizeHint(self):
        return QSize(self.SIZE, self.SIZE)


class NextActionIcon(QLabel):
    STYLESHEET = 'background: transparent; padding-left: 1px'
    SIZE = 16

    def __init__(self, cue, *args):
        super().__init__(*args)
        self.setStyleSheet(self.STYLESHEET)

        self.cue = cue
        self.cue.changed('next_action').connect(
            self._update_icon, Connection.QtQueued)

        self._update_icon(self.cue.next_action)

    def _update_icon(self, next_action):
        next_action = CueNextAction(next_action)
        pixmap = pixmap_from_icon('', self.SIZE)

        if next_action == CueNextAction.AutoNext:
            pixmap = pixmap_from_icon('auto-next', self.SIZE)
            self.setToolTip(CueNextAction.AutoNext.value)
        elif next_action == CueNextAction.AutoFollow:
            pixmap = pixmap_from_icon('auto-follow', self.SIZE)
            self.setToolTip(CueNextAction.AutoFollow.value)
        else:
            self.setToolTip('')

        self.setPixmap(pixmap)

    def sizeHint(self):
        return QSize(self.SIZE + 2, self.SIZE)


class TimeWidget(QProgressBar):

    def __init__(self, cue, *args):
        super().__init__(*args)
        self.setObjectName('ListTimeWidget')
        self.setValue(0)
        self.setTextVisible(True)

        self.show_zero_duration = False
        self.accurate_time = True
        self.cue = cue

    def _update_time(self, time):
        self.setValue(time)
        self.setFormat(strtime(time, accurate=self.accurate_time))

    def _update_duration(self, duration):
        if duration > 0 or self.show_zero_duration:
            # Display as disabled if duration < 0
            self.setEnabled(duration > 0)
            self.setTextVisible(True)
            self.setFormat(strtime(duration, accurate=self.accurate_time))
            # Avoid settings min and max to 0, or the the bar go in busy state
            self.setRange(0 if duration > 0 else -1, duration)
        else:
            self.setTextVisible(False)

    def _update_style(self, state):
        self.setProperty('state', state)
        self.style().unpolish(self)
        self.style().polish(self)
        self.update()

    def _running(self):
        self._update_style('running')

    def _pause(self):
        self._update_style('pause')
        self._update_time(self.value())

    def _stop(self):
        self._update_style('stop')
        self.setValue(self.minimum())

    def _error(self):
        self._update_style('error')
        self.setValue(self.minimum())


class CueTimeWidget(TimeWidget):

    def __init__(self, cue, *args):
        super().__init__(cue, *args)

        self.cue.interrupted.connect(self._stop, Connection.QtQueued)
        self.cue.started.connect(self._running, Connection.QtQueued)
        self.cue.stopped.connect(self._stop, Connection.QtQueued)
        self.cue.paused.connect(self._pause, Connection.QtQueued)
        self.cue.error.connect(self._error, Connection.QtQueued)
        self.cue.end.connect(self._stop, Connection.QtQueued)
        self.cue.changed('duration').connect(
            self._update_duration, Connection.QtQueued)

        self.cue_time = CueTime(self.cue)
        self.cue_time.notify.connect(self._update_time, Connection.QtQueued)

        if cue.state & CueState.Running:
            self._running()
        elif cue.state & CueState.Pause:
            self._pause()
        elif cue.state & CueState.Error:
            self._error()
        else:
            self._stop()

    def _stop(self):
        super()._stop()
        self._update_duration(self.cue.duration)


class PreWaitWidget(TimeWidget):

    def __init__(self, cue, *args):
        super().__init__(cue, *args)
        self.show_zero_duration = True

        self.cue.prewait_start.connect(self._running, Connection.QtQueued)
        self.cue.prewait_stopped.connect(self._stop, Connection.QtQueued)
        self.cue.prewait_paused.connect(self._pause, Connection.QtQueued)
        self.cue.prewait_ended.connect(self._stop, Connection.QtQueued)
        self.cue.changed('pre_wait').connect(
            self._update_duration, Connection.QtQueued)

        self._update_duration(self.cue.pre_wait)

        self.wait_time = CueWaitTime(self.cue, mode=CueWaitTime.Mode.Pre)
        self.wait_time.notify.connect(self._update_time, Connection.QtQueued)

    def _update_duration(self, duration):
        # The wait time is in seconds, we need milliseconds
        super()._update_duration(duration * 1000)

    def _stop(self):
        super()._stop()
        self._update_duration(self.cue.pre_wait)


class PostWaitWidget(TimeWidget):

    def __init__(self, cue, *args):
        super().__init__(cue, *args)
        self.show_zero_duration = True

        self.cue.changed('next_action').connect(
            self._next_action_changed, Connection.QtQueued)

        self.wait_time = CueWaitTime(self.cue, mode=CueWaitTime.Mode.Post)
        self.cue_time = CueTime(self.cue)

        self._next_action_changed(self.cue.next_action)

    def _update_duration(self, duration):
        if self.cue.next_action != CueNextAction.AutoFollow.value:
            # The wait time is in seconds, we need milliseconds
            duration *= 1000

        super()._update_duration(duration)

    def _next_action_changed(self, next_action):
        self.cue.postwait_start.disconnect(self._running)
        self.cue.postwait_stopped.disconnect(self._stop)
        self.cue.postwait_paused.disconnect(self._pause)
        self.cue.postwait_ended.disconnect(self._stop)

        self.cue.interrupted.disconnect(self._stop)
        self.cue.started.disconnect(self._running)
        self.cue.stopped.disconnect(self._stop)
        self.cue.paused.disconnect(self._pause)
        self.cue.error.disconnect(self._stop)
        self.cue.end.disconnect(self._stop)

        self.cue_time.notify.disconnect(self._update_time)
        self.wait_time.notify.disconnect(self._update_time)

        self.cue.changed('post_wait').disconnect(self._update_duration)
        self.cue.changed('duration').disconnect(self._update_duration)

        if next_action == CueNextAction.AutoFollow.value:
            self.cue.interrupted.connect(self._stop, Connection.QtQueued)
            self.cue.started.connect(self._running, Connection.QtQueued)
            self.cue.stopped.connect(self._stop, Connection.QtQueued)
            self.cue.paused.connect(self._pause, Connection.QtQueued)
            self.cue.error.connect(self._stop, Connection.QtQueued)
            self.cue.end.connect(self._stop, Connection.QtQueued)
            self.cue.changed('duration').connect(
                self._update_duration, Connection.QtQueued)

            self.cue_time.notify.connect(self._update_time, Connection.QtQueued)
            self._update_duration(self.cue.duration)
        else:
            self.cue.postwait_start.connect(self._running, Connection.QtQueued)
            self.cue.postwait_stopped.connect(self._stop, Connection.QtQueued)
            self.cue.postwait_paused.connect(self._pause, Connection.QtQueued)
            self.cue.postwait_ended.connect(self._stop, Connection.QtQueued)
            self.cue.changed('post_wait').connect(
                self._update_duration, Connection.QtQueued)

            self.wait_time.notify.connect(
                self._update_time, Connection.QtQueued)
            self._update_duration(self.cue.post_wait)

    def _stop(self):
        super()._stop()

        if self.cue.next_action == CueNextAction.AutoFollow.value:
            self._update_duration(self.cue.duration)
        else:
            self._update_duration(self.cue.post_wait)


# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt, QTime
from PyQt5.QtWidgets import QGroupBox, QComboBox, QVBoxLayout, QTimeEdit

from lisp.modules.gst_backend.elements.preset_src import PresetSrc
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class PresetSrcSettings(SettingsPage):
    ELEMENT = PresetSrc
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.functionGroup = QGroupBox(self)
        self.functionGroup.setTitle(translate('PresetSrcSettings', 'Presets'))
        self.functionGroup.setLayout(QVBoxLayout())
        self.layout().addWidget(self.functionGroup)

        self.functionCombo = QComboBox(self.functionGroup)
        self.functionGroup.layout().addWidget(self.functionCombo)

        for function in sorted(PresetSrc.PRESETS.keys()):
            self.functionCombo.addItem(function)

        self.functionDuration = QTimeEdit(self.functionGroup)
        self.functionDuration.setDisplayFormat('HH.mm.ss.zzz')
        self.functionGroup.layout().addWidget(self.functionDuration)

    def enable_check(self, enabled):
        self.functionGroup.setCheckable(enabled)
        self.functionGroup.setChecked(False)

    def get_settings(self):
        if not (self.functionGroup.isCheckable() and not self.functionGroup.isChecked()):
            return {'preset': self.functionCombo.currentText(),
                    'duration': self.functionDuration.time().
                        msecsSinceStartOfDay()}

        return {}

    def load_settings(self, settings):
        self.functionCombo.setCurrentText(settings.get('preset', ''))
        self.functionDuration.setTime(
            QTime.fromMSecsSinceStartOfDay(settings.get('duration', 0)))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import os
from configparser import ConfigParser
from shutil import copyfile

DEFAULT_CFG_PATH = os.path.join(os.path.dirname(__file__), '../default.cfg')
CFG_DIR = os.path.expanduser("~") + '/.linux_show_player'
CFG_PATH = CFG_DIR + '/config.cfg'

config = ConfigParser()


def load_config():
    # Check if the current user configuration is up-to-date
    check_user_config()
    # Read the user configuration
    config.read(CFG_PATH)


def write_config():
    with open(CFG_PATH, 'w') as f:
        config.write(f)


def check_user_config():
    update = True

    if not os.path.exists(CFG_DIR):
        os.makedirs(CFG_DIR)
    elif os.path.exists(CFG_PATH):
        default = ConfigParser()
        default.read(DEFAULT_CFG_PATH)

        current = ConfigParser()
        current.read(CFG_PATH)

        current_version = current['Version']['Number']
        update = current_version != default['Version']['Number']

        if update:
            copyfile(CFG_PATH, CFG_PATH + '.old')
            print('Configuration file backup: {}.old'.format(CFG_PATH))

    if update:
        copyfile(DEFAULT_CFG_PATH, CFG_PATH)
        print('Create new configuration file: {}'.format(CFG_PATH))


def config_to_dict():
    conf_dict = {}

    for section in config.keys():
        conf_dict[section] = {}
        for option in config[section].keys():
            conf_dict[section][option] = config[section][option]

    return conf_dict


def update_config_from_dict(conf):
    for section in conf.keys():
        for option in conf[section].keys():
            config[section][option] = conf[section][option]

    write_config()

# Load configuration
load_config()


# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from lisp.core.signal import Signal
from lisp.modules.midi.midi_common import MIDICommon
from lisp.modules.midi.midi_utils import mido_backend, mido_port_name


class MIDIInput(MIDICommon):
    def __init__(self, port_name='AppDefault'):
        super().__init__(port_name=port_name)

        self.alternate_mode = False
        self.new_message = Signal()
        self.new_message_alt = Signal()

    def open(self):
        port_name = mido_port_name(self._port_name, 'I')
        self._port = mido_backend().open_input(name=port_name,
                                               callback=self.__new_message)

    def __new_message(self, message):
        if self.alternate_mode:
            self.new_message_alt.emit(message)
        else:
            self.new_message.emit(message)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from threading import Event

from lisp.core.decorators import locked_method
from lisp.core.fade_functions import ntime, FadeInType, FadeOutType
from lisp.core.util import rsetattr, rgetattr


class Fader:
    """Allow to perform fades on "generic" objects attributes.

     * Fades have a resolution of `1-hundredth-of-second`
     * To be able to fade correctly the attribute must be numeric, if not, the
       `fade` function will fail
     * When fading, the `fade` function cannot be entered by other threads,
       if this happens the function will simply return immediately
     * To execute a fader, the `prepare` function must be called first,
       this will also stop the fader
     * After calling `prepare` the fader is considered as running
     * The `stop` function wait until all "pending" target changes are applied
     * Changing the target will also stop the fader
    """

    def __init__(self, target, attribute):
        """
        :param target: The target object
        :type target: object
        :param attribute: The target attribute (name) to be faded
        :type attribute: str
        """
        self._time = 0  # current fade time in hundredths-of-seconds
        self._target = target
        self._attribute = attribute

        self._is_ready = Event()
        self._is_ready.set()
        self._running = Event()
        self._running.set()
        self._pause = Event()
        self._pause.set()

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target):
        self.stop()
        self._target = target

    @property
    def attribute(self):
        return self._attribute

    @attribute.setter
    def attribute(self, target_property):
        self.stop()
        self._attribute = target_property

    def prepare(self):
        self.stop()

        self._running.clear()
        self._is_ready.clear()

    @locked_method(blocking=False)
    def fade(self, duration, to_value, fade_type):
        """
        :param duration: How much the fade should be long (in seconds)
        :type duration: float
        :param to_value: The value to reach
        :type to_value: float
        :param fade_type: The fade type
        :type fade_type: FadeInType | FadeOutType

        :return: False if the fade as been interrupted, True otherwise
        :rtype: bool
        """
        if duration <= 0:
            return

        if not isinstance(fade_type, (FadeInType, FadeOutType)):
            raise AttributeError(
                'fade_type must be one of FadeInType or FadeOutType member,'
                'not {}'.format(fade_type.__class__.__name__))

        try:
            self._time = 0
            begin = 0
            functor = fade_type.value
            duration = int(duration * 100)
            base_value = rgetattr(self._target, self._attribute)
            value_diff = to_value - base_value

            if value_diff == 0:
                return

            while self._time <= duration and not self._running.is_set():
                rsetattr(self._target,
                         self._attribute,
                         functor(ntime(self._time, begin, duration),
                                 value_diff,
                                 base_value))

                self._time += 1
                self._running.wait(0.01)
                self._pause.wait()
        finally:
            interrupted = self._running.is_set()
            self._running.set()
            self._is_ready.set()
            self._time = 0

        return not interrupted

    def stop(self):
        if not self._running.is_set() or not self._pause.is_set():
            self._running.set()
            self._pause.set()
            self._is_ready.wait()

    def pause(self):
        if self.is_running():
            self._pause.clear()

    def restart(self):
        self._pause.set()

    def is_paused(self):
        return not self._pause.is_set()

    def is_running(self):
        return not self._running.is_set() and not self.is_paused()

    def current_time(self):
        # Return the time in millisecond
        return self._time * 10



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from enum import IntEnum
from weakref import WeakValueDictionary

from lisp.core.clock import Clock_100
from lisp.core.decorators import locked_method
from lisp.core.signal import Connection, Signal
from lisp.cues.cue import CueState


class MetaCueTime(type):
    """Allow "caching" of CueTime(s) objects."""

    __Instances = WeakValueDictionary()

    @locked_method
    def __call__(cls, cue, *args, **kwargs):
        instance = MetaCueTime.__Instances.get((cls, cue.id))
        if instance is None:
            instance = super().__call__(cue, *args, **kwargs)
            MetaCueTime.__Instances[(cls, cue.id)] = instance

        return instance


class CueTime(metaclass=MetaCueTime):
    """Provide timing for a Cue.

    Once created the notify signal provide timing for the given cue.
    The current time is queried using `Cue.current_time()`.

    .. note::
        The notify signal is emitted only when the cue is running.
    """
    _Clock = Clock_100

    def __init__(self, cue):
        self.notify = Signal()

        self._clock = self._Clock
        self._active = False
        self._cue = cue
        self._cue.changed('duration').connect(self.__init)

        self.__init()

    def __init(self, *args):
        if self._cue.duration > 0 and not self._active:
            # Cue "status" signals
            self._cue.started.connect(self.start, Connection.QtQueued)
            self._cue.paused.connect(self.stop, Connection.QtQueued)
            self._cue.stopped.connect(self.stop, Connection.QtQueued)
            self._cue.end.connect(self.stop, Connection.QtQueued)
            self._cue.error.connect(self.stop, Connection.QtQueued)

            if self._cue.state & CueState.Running:
                self.start()
            self._active = True
        elif self._cue.duration < 0 and self._active:
            self._cue.started.disconnect(self.start)
            self._cue.paused.disconnect(self.stop)
            self._cue.stopped.disconnect(self.stop)
            self._cue.end.disconnect(self.stop)
            self._cue.error.disconnect(self.stop)

            self.stop()
            self._active = False

    def __notify(self):
        """Notify the cue current-time"""
        if self._cue.state & (CueState.Running ^ CueState.Pause):
            self.notify.emit(self._cue.current_time())

    def start(self):
        self._clock.add_callback(self.__notify)

    def stop(self):
        try:
            self._clock.remove_callback(self.__notify)
        except Exception:
            # TODO: catch only the exception when the callback is not registered
            pass


class CueWaitTime:
    """Provide timing for Cue pre/post waits.

    Once created the notify signal provide timing for the specified wait for the
    given cue.
    The time since the wait start is calculated internally using the :mod:`time`
    module functions.
    """

    class Mode(IntEnum):
        Pre = 0
        Post = 1

    def __init__(self, cue, mode=Mode.Pre):
        self.notify = Signal()

        self._clock = Clock_100
        self._start_time = 0
        self._last = 0
        self._cue = cue
        self._mode = mode

        if self._mode == CueWaitTime.Mode.Pre:
            self._cue.prewait_ended.connect(self.stop, Connection.QtQueued)
            self._cue.prewait_start.connect(self.start, Connection.QtQueued)
            self._cue.prewait_paused.connect(self.stop, Connection.QtQueued)
            self._cue.prewait_stopped.connect(self.stop, Connection.QtQueued)
        elif self._mode == CueWaitTime.Mode.Post:
            self._cue.postwait_ended.connect(self.stop, Connection.QtQueued)
            self._cue.postwait_start.connect(self.start, Connection.QtQueued)
            self._cue.postwait_paused.connect(self.stop, Connection.QtQueued)
            self._cue.postwait_stopped.connect(self.stop, Connection.QtQueued)

    def __notify(self):
        if self._mode == CueWaitTime.Mode.Pre:
            self.notify.emit(int(self._cue.prewait_time() * 100) * 10)
        else:
            self.notify.emit(int(self._cue.postwait_time() * 100) * 10)

    def start(self):
        self._clock.add_callback(self.__notify)

    def stop(self):
        try:
            self._clock.remove_callback(self.__notify)
        except Exception:
            # TODO: catch only the exception when the callback is not registered
            pass



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABCMeta
from copy import deepcopy

from lisp.core.signal import Signal
from lisp.core.util import subclasses


class Property:
    """Descriptor to be used in HasProperties subclasses to define properties.

    .. warning::
        To be able to save properties into a session, the stored value
        MUST be JSON-serializable.

    .. warning::
        If extended any subclass *MUST*:
        1) if the __get__ method receive a None instance return self;
        2) if the __get__ is called while the value is not set, set it with a
           safe copy of 'default' and return.
        3) After the value is changed call the __changed__ method.
    """

    def __init__(self, default=None):
        self.name = 'unnamed_property'
        self.default = default

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        elif self.name not in instance.__dict__:
            instance.__dict__[self.name] = deepcopy(self.default)

        return instance.__dict__.get(self.name)

    def __set__(self, instance, value):
        if instance is not None:
            # Only change the value if different
            if value != instance.__dict__.get(self.name, self.default):
                instance.__dict__[self.name] = value
                self.__changed__(instance, value)

    def changed(self, instance):
        if instance is not None:
            value = self.__get__(instance)
            return value != self.default, value

        return False, self.default

    def __changed__(self, instance, value):
        instance.property_changed.emit(instance, self.name, value)
        # Get the related signal
        property_signal = instance.changed_signals.get(self.name)
        if property_signal is not None:
            property_signal.emit(value)


class WriteOnceProperty(Property):
    """Property that can be modified only once.

    Obviously this is not really "write-once", but if used as normal attribute
    will ignore any change when the stored value is different from default.
    """

    def __set__(self, instance, value):
        if self.__get__(instance) == self.default:
            super().__set__(instance, value)


class NestedProperties(Property):
    """Simplify retrieving the properties of nested HasProperties objects.

    The goal is to avoid the reimplementation of HasProperties.properties()
    and HasProperties.update_properties().

    ..note::
        When getting or setting a single property of the nested object is better
        to access it directly instead that using the nested-property.
    """

    def __init__(self, provider_name, default=None):
        super().__init__(default=default)
        self.provider_name = provider_name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        else:
            provider = instance.__dict__.get(self.provider_name)
            if isinstance(provider, HasProperties):
                return provider.properties()

    def __set__(self, instance, value):
        if instance is not None:
            provider = instance.__dict__.get(self.provider_name)
            if isinstance(provider, HasProperties):
                provider.update_properties(value)
                self.__changed__(instance, value)

    def changed(self, instance):
        if instance is not None:
            provider = instance.__dict__.get(self.provider_name)
            if isinstance(provider, HasProperties):
                properties = provider.properties(only_changed=True)
                # If no properties is changed (empty dict) return false
                return bool(properties), properties

        return False, {}


class HasPropertiesMeta(ABCMeta):
    """Metaclass for HasProperties classes.

    This metaclass manage the 'propagation' of the properties in all subclasses,
    this process involves overwriting __properties__ with a set containing all
    the properties names.

    ..note::
        This metaclass is derived form :class:`abc.ABCMeta`, so abstract
        classes can be created without using an intermediate metaclass
    """

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Use a set for avoiding repetitions
        cls.__properties__ = set()

        # Populate with all the properties
        for name, attribute in vars(cls).items():
            if isinstance(attribute, Property):
                cls.__properties__.add(name)
                attribute.name = name

        for base in cls.__bases__:
            cls.__properties__.update(getattr(base, '__properties__', ()))


class HasProperties(metaclass=HasPropertiesMeta):
    """Base class providing a simple way to mange object properties.

    Using the Property descriptor, subclasses, can specify a set of
    properties, that can be easily retrieved and updated via :func:`properties`
    and :func:`update_properties`.

    .. Usage::

        class MyClass(HasProperties):
            prop1 = Property(default=100)
            prop2 = Property()
    """

    __properties__ = set()

    def __init__(self):
        self.property_changed = Signal()
        #: Emitted after property change (self, name, value)

        self.changed_signals = {}
        """Contains signals that are emitted after the associated property is
        changed, the signal are create only when requested the first time.
        """

    @classmethod
    def register_property(cls, name, prop):
        """Register a new property with the given name.

        :param str name: Property name
        :param BaseProperty prop: The property
        """
        if name not in cls.__properties__:
            prop.name = name
            setattr(cls, name, prop)
            cls.__properties__.add(name)

            for subclass in subclasses(cls):
                subclass.__properties__.add(name)

    def changed(self, property_name):
        """
        :param property_name: The property name
        :return: The property change-signal
        :rtype: Signal
        """
        if property_name not in self.__class__.__properties__:
            raise ValueError('no property "{0}" found'.format(property_name))

        signal = self.changed_signals.get(property_name, None)

        if signal is None:
            signal = Signal()
            self.changed_signals[property_name] = signal

        return signal

    def properties(self, only_changed=False):
        """
        :param only_changed: when True only "changed" properties are collected
        :type only_changed: bool

        :return: The properties as a dictionary {name: value}
        :rtype: dict
        """
        if only_changed:
            properties = {}
            for name in self.__properties__:
                changed, value = getattr(self.__class__, name).changed(self)
                if changed:
                    properties[name] = value

            return properties

        return {name: getattr(self, name) for name in self.__properties__}

    @classmethod
    def properties_defaults(cls):
        """
        :return: The default properties as a dictionary {name: default_value}
        :rtype: dict
        """
        return {name: getattr(cls, name).default for name in cls.__properties__}

    def update_properties(self, properties):
        """Set the given properties.

        :param properties: The element properties
        :type properties: dict
        """
        for name, value in properties.items():
            if name in self.__properties__:
                setattr(self, name, value)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QT_TRANSLATE_NOOP

from lisp.backend.media_element import ElementType, MediaType
from lisp.core.has_properties import Property
from lisp.modules.gst_backend.gi_repository import Gst
from lisp.modules.gst_backend.gst_element import GstMediaElement


class Speed(GstMediaElement):
    ElementType = ElementType.Plugin
    MediaType = MediaType.Audio
    Name = QT_TRANSLATE_NOOP('MediaElementName', 'Speed')

    speed = Property(default=1.0)

    def __init__(self, pipeline):
        super().__init__()

        self.pipeline = pipeline
        self.scale_tempo = Gst.ElementFactory.make("scaletempo", None)
        self.audio_convert = Gst.ElementFactory.make("audioconvert", None)

        self.pipeline.add(self.scale_tempo)
        self.pipeline.add(self.audio_convert)

        self.scale_tempo.link(self.audio_convert)

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        self._handler = bus.connect("message", self.__on_message)

        self._old_speed = self.speed
        self.changed('speed').connect(self.__prepare_speed)

    def __prepare_speed(self, value):
        if self._old_speed != value:
            self._old_speed = value

            if self.pipeline.current_state == Gst.State.PLAYING:
                self.__change_speed()

    def sink(self):
        return self.scale_tempo

    def src(self):
        return self.audio_convert

    def dispose(self):
        bus = self.pipeline.get_bus()
        bus.remove_signal_watch()
        bus.disconnect(self._handler)

    def __on_message(self, bus, message):
        if (message.type == Gst.MessageType.STATE_CHANGED and
                    message.src == self.scale_tempo and
                    message.parse_state_changed()[1] == Gst.State.PLAYING):
            self.__change_speed()

    def __change_speed(self):
        current_position = self.scale_tempo.query_position(Gst.Format.TIME)[1]

        self.scale_tempo.seek(self.speed,
                              Gst.Format.TIME,
                              Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE,
                              Gst.SeekType.SET,
                              current_position,
                              Gst.SeekType.NONE,
                              0)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtGui import QLinearGradient, QColor, QPainter
from PyQt5.QtWidgets import QWidget

from lisp.core.configuration import config
from lisp.core.decorators import suppress_exceptions


class QDbMeter(QWidget):
    DB_MIN = int(config["DbMeter"]["dbMin"])
    DB_MAX = int(config["DbMeter"]["dbMax"])
    DB_CLIP = int(config["DbMeter"]["dbClip"])

    def __init__(self, parent):
        super().__init__(parent)

        db_range = abs(self.DB_MIN - self.DB_MAX)
        yellow = abs(self.DB_MIN + 20) / db_range  # -20 db
        red = abs(self.DB_MIN) / db_range          # 0 db

        self.grad = QLinearGradient()
        self.grad.setColorAt(0, QColor(0, 255, 0))            # Green
        self.grad.setColorAt(yellow, QColor(255, 255, 0))     # Yellow
        self.grad.setColorAt(red, QColor(255, 0, 0))          # Red

        self.reset()

    def reset(self):
        self.peaks = [self.DB_MIN, self.DB_MIN]
        self.rmss = [self.DB_MIN, self.DB_MIN]
        self.decPeak = [self.DB_MIN, self.DB_MIN]
        self.clipping = {}
        self.repaint()

    def plot(self, peaks, rms, decPeak):
        self.peaks = peaks
        self.rmss = rms
        self.decPeak = decPeak

        self.repaint()

    @suppress_exceptions
    def paintEvent(self, e):
        if not self.visibleRegion().isEmpty():
            # Stretch factor
            mul = (self.height() - 4)
            mul /= (self.DB_MAX - self.DB_MIN)

            peaks = []
            for n, peak in enumerate(self.peaks):
                if peak > self.DB_CLIP:
                    self.clipping[n] = True

                if peak < self.DB_MIN:
                    peak = self.DB_MIN
                elif peak > self.DB_MAX:
                    peak = self.DB_MAX

                peaks.append(round((peak - self.DB_MIN) * mul))

            rmss = []
            for n, rms in enumerate(self.rmss):
                if rms < self.DB_MIN:
                    rms = self.DB_MIN
                elif rms > self.DB_MAX:
                    rms = self.DB_MAX

                rmss.append(round((rms - self.DB_MIN) * mul))

            dPeaks = []
            for dPeak in self.decPeak:
                if dPeak < self.DB_MIN:
                    dPeak = self.DB_MIN
                elif dPeak > self.DB_MAX:
                    dPeak = self.DB_MAX

                dPeaks.append(round((dPeak - self.DB_MIN) * mul))

            qp = QPainter()
            qp.begin(self)
            qp.setBrush(QColor(0, 0, 0, 0))

            xpos = 0
            xdim = self.width() / len(peaks)

            for n, (peak, rms, dPeak) in enumerate(zip(peaks, rmss, dPeaks)):
                # Maximum "peak-rect" size
                maxRect = QtCore.QRect(xpos, self.height() - 2, xdim - 2,
                                       2 - self.height())

                # Set QLinearGradient start and final-stop position
                self.grad.setStart(maxRect.topLeft())
                self.grad.setFinalStop(maxRect.bottomRight())

                # Draw peak (audio peak in dB)
                rect = QtCore.QRect(xpos, self.height() - 2, xdim - 2, -peak)
                qp.setOpacity(0.6)
                qp.fillRect(rect, self.grad)
                qp.setOpacity(1.0)

                # Draw rms (in db)
                rect = QtCore.QRect(xpos, self.height() - 2, xdim - 2, -rms)
                qp.fillRect(rect, self.grad)

                # Draw decay peak
                decRect = QtCore.QRect(xpos, (self.height() - 3) - dPeak,
                                       xdim - 2, 2)
                qp.fillRect(decRect, self.grad)

                # Draw Borders
                if self.clipping.get(n, False):
                    qp.setPen(QColor(200, 0, 0))
                else:
                    qp.setPen(QColor(100, 100, 100))

                qp.drawRect(maxRect)

                xpos += xdim

            qp.end()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import deque

from lisp.core import configuration as cfg
from lisp.core.action import Action
from lisp.core.signal import Signal


class ActionsHandler:
    """Provide a classic undo/redo mechanism based on stacks."""

    MaxStackSize = int(cfg.config['Actions']['MaxStackSize'])

    def __init__(self):
        super().__init__()

        self.action_done = Signal()
        self.action_undone = Signal()
        self.action_redone = Signal()

        self._undo = deque()
        self._redo = deque()
        if self.MaxStackSize > 0:
            self._undo.maxlen = self.MaxStackSize
            self._redo.maxlen = self.MaxStackSize

        self._saved_action = None

    def clear(self):
        """Clear the `undo` and `redo` stacks."""
        self._undo.clear()
        self._redo.clear()
        self._saved_action = None

    def do_action(self, action: Action):
        """Execute the action, and add it the `undo` stack.

        When an action is executed:
         * is logged
         * is appended to the `undo` stack
         * the `redo` stack is cleared to maintain consistency
         * the `action_done` signal is emitted
        """
        action.do()

        self._logging(action, 'Last action: ')
        self._undo.append(action)
        # Clean the redo stack for maintain consistency
        self._redo.clear()

        self.action_done.emit(action)

    def undo_action(self):
        """Undo the last executed action, and add it to the `redo` stack.

        When an action is undone:
         * is removed from the `undo` stack
         * is logged
         * is appended to the `redo` stack
         * the signal `action_undone` is emitted
        """
        if self._undo:
            action = self._undo.pop()
            action.undo()

            self._logging(action, 'Undo: ')
            self._redo.append(action)

            self.action_undone.emit(action)

    def redo_action(self):
        """Redo the last undone action, and add it back to the `undo` stack.

        When an action is redone:
         * is remove from the `redo` stack
         * is logged
         * is added to the `undo` stack
         * the `action_redone` signal is emitted
        """
        if self._redo:
            action = self._redo.pop()
            action.redo()

            self._logging(action, 'Redo: ')
            self._undo.append(action)

            self.action_redone.emit(action)

    def set_saved(self):
        """Set the action at the _top_ of the `undo` stack as `save-point`."""
        if self._undo:
            self._saved_action = self._undo[-1]

    def is_saved(self) -> bool:
        """Return True if the action at the _top_ of the `undo` stack is the
        `save-point`.
        """
        if self._undo:
            return self._undo[-1] is self._saved_action
        else:
            return True

    @staticmethod
    def _logging(action: Action, pre: str):
        message = action.log()
        if message.strip() == '':
            message = type(action).__name__

        logging.info(pre + message)


MainActionsHandler = ActionsHandler()  # "global" action-handler



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt, QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QTabWidget, QAction, QInputDialog, qApp, \
    QMessageBox

from lisp.core.configuration import config
from lisp.core.signal import Connection
from lisp.cues.cue import Cue
from lisp.cues.cue_factory import CueFactory
from lisp.cues.media_cue import MediaCue
from lisp.layouts.cart_layout.cart_layout_settings import CartLayoutSettings
from lisp.layouts.cart_layout.cue_cart_model import CueCartModel
from lisp.layouts.cart_layout.cue_widget import CueWidget
from lisp.layouts.cart_layout.page_widget import PageWidget
from lisp.layouts.cue_layout import CueLayout
from lisp.ui.mainwindow import MainWindow
from lisp.ui.settings.app_settings import AppSettings
from lisp.ui.settings.cue_settings import CueSettingsRegistry
from lisp.ui.settings.pages.cue_appearance import Appearance
from lisp.ui.settings.pages.cue_general import CueGeneralSettings
from lisp.ui.settings.pages.media_cue_settings import MediaCueSettings
from lisp.ui.ui_utils import translate

AppSettings.register_settings_widget(CartLayoutSettings)


class CartLayout(QTabWidget, CueLayout):
    NAME = 'Cart Layout'
    DESCRIPTION = translate('LayoutDescription',
                            'Organize cues in grid like pages')
    DETAILS = [
        QT_TRANSLATE_NOOP('LayoutDetails', 'Click a cue to run it'),
        QT_TRANSLATE_NOOP('LayoutDetails', 'SHIFT + Click to edit a cue'),
        QT_TRANSLATE_NOOP('LayoutDetails', 'CTRL + Click to select a cue'),
        QT_TRANSLATE_NOOP('LayoutDetails',
                          'To copy cues drag them while pressing CTRL'),
        QT_TRANSLATE_NOOP('LayoutDetails',
                          'To copy cues drag them while pressing SHIFT')
    ]

    def __init__(self, cue_model, **kwargs):
        super().__init__(cue_model=cue_model, **kwargs)
        self.tabBar().setObjectName('CartTabBar')

        self.__columns = int(config['CartLayout']['GridColumns'])
        self.__rows = int(config['CartLayout']['GridRows'])
        self.__pages = []
        self.__context_widget = None

        self._show_seek = config['CartLayout'].getboolean('ShowSeek')
        self._show_dbmeter = config['CartLayout'].getboolean('ShowDbMeters')
        self._show_volume = config['CartLayout'].getboolean('ShowVolume')
        self._accurate_timing = config['CartLayout'].getboolean('ShowAccurate')
        self._countdown_mode = config['CartLayout'].getboolean('CountDown')
        self._auto_add_page = config['CartLayout'].getboolean('AutoAddPage')

        self._model_adapter = CueCartModel(cue_model, self.__rows, self.__columns)
        self._model_adapter.item_added.connect(self.__cue_added, Connection.QtQueued)
        self._model_adapter.item_removed.connect(self.__cue_removed, Connection.QtQueued)
        self._model_adapter.item_moved.connect(self.__cue_moved, Connection.QtQueued)
        self._model_adapter.model_reset.connect(self.__model_reset)

        # Add layout-specific menus
        self.new_page_action = QAction(self)
        self.new_page_action.triggered.connect(self.add_page)
        self.new_pages_action = QAction(self)
        self.new_pages_action.triggered.connect(self.add_pages)
        self.rm_current_page_action = QAction(self)
        self.rm_current_page_action.triggered.connect(self.remove_current_page)

        self.countdown_mode = QAction(self)
        self.countdown_mode.setCheckable(True)
        self.countdown_mode.setChecked(self._countdown_mode)
        self.countdown_mode.triggered.connect(self.set_countdown_mode)

        self.show_seek_action = QAction(self)
        self.show_seek_action.setCheckable(True)
        self.show_seek_action.setChecked(self._show_seek)
        self.show_seek_action.triggered.connect(self.set_seek_visible)

        self.show_dbmeter_action = QAction(self)
        self.show_dbmeter_action.setCheckable(True)
        self.show_dbmeter_action.setChecked(self._show_dbmeter)
        self.show_dbmeter_action.triggered.connect(self.set_dbmeter_visible)

        self.show_volume_action = QAction(self)
        self.show_volume_action.setCheckable(True)
        self.show_volume_action.setChecked(self._show_volume)
        self.show_volume_action.triggered.connect(self.set_volume_visible)

        self.show_accurate_action = QAction(self)
        self.show_accurate_action.setCheckable(True)
        self.show_accurate_action.setChecked(self._accurate_timing)
        self.show_accurate_action.triggered.connect(self.set_accurate)

        layoutMenu = MainWindow().menuLayout
        layoutMenu.addAction(self.new_page_action)
        layoutMenu.addAction(self.new_pages_action)
        layoutMenu.addAction(self.rm_current_page_action)
        layoutMenu.addSeparator()
        layoutMenu.addAction(self.countdown_mode)
        layoutMenu.addAction(self.show_seek_action)
        layoutMenu.addAction(self.show_dbmeter_action)
        layoutMenu.addAction(self.show_volume_action)
        layoutMenu.addAction(self.show_accurate_action)

        # TODO: maybe can be moved outside the layout
        # Add cue preferences widgets
        CueSettingsRegistry().add_item(CueGeneralSettings, Cue)
        CueSettingsRegistry().add_item(MediaCueSettings, MediaCue)
        CueSettingsRegistry().add_item(Appearance)

        # Cue(s) context-menu actions
        self.edit_action = QAction(self)
        self.edit_action.triggered.connect(self._edit_cue_action)
        self.cm_registry.add_item(self.edit_action)

        self.sep1 = self.cm_registry.add_separator()

        self.remove_action = QAction(self)
        self.remove_action.triggered.connect(self._remove_cue_action)
        self.cm_registry.add_item(self.remove_action)

        self.select_action = QAction(self)
        self.select_action.triggered.connect(self.select_context_cue)
        self.cm_registry.add_item(self.select_action)

        self.sep2 = self.cm_registry.add_separator(MediaCue)

        # MediaCue(s) context-menu actions
        self.play_action = QAction(self)
        self.play_action.triggered.connect(self._play_context_cue)
        self.cm_registry.add_item(self.play_action, MediaCue)

        self.pause_action = QAction(self)
        self.pause_action.triggered.connect(self._pause_context_cue)
        self.cm_registry.add_item(self.pause_action, MediaCue)

        self.stop_action = QAction(self)
        self.stop_action.triggered.connect(self._stop_context_cue)
        self.cm_registry.add_item(self.stop_action, MediaCue)

        self.reset_volume_action = QAction(self)
        self.reset_volume_action.triggered.connect(self._reset_cue_volume)
        self.cm_registry.add_item(self.reset_volume_action, MediaCue)

        self.setAcceptDrops(True)
        self.retranslateUi()

        self.add_page()

    def retranslateUi(self):
        self.new_page_action.setText(translate('CartLayout', 'Add page'))
        self.new_pages_action.setText(translate('CartLayout', 'Add pages'))
        self.rm_current_page_action.setText(
            translate('CartLayout', 'Remove current page'))
        self.countdown_mode.setText(translate('CartLayout', 'Countdown mode'))
        self.show_seek_action.setText(translate('CartLayout', 'Show seek-bars'))
        self.show_dbmeter_action.setText(
            translate('CartLayout', 'Show dB-meters'))
        self.show_volume_action.setText(translate('CartLayout', 'Show volume'))
        self.show_accurate_action.setText(
            translate('CartLayout', 'Show accurate time'))

        self.edit_action.setText(translate('CartLayout', 'Edit cue'))
        self.remove_action.setText(translate('CartLayout', 'Remove'))
        self.select_action.setText(translate('CartLayout', 'Select'))
        self.play_action.setText(translate('CartLayout', 'Play'))
        self.pause_action.setText(translate('CartLayout', 'Pause'))
        self.stop_action.setText(translate('CartLayout', 'Stop'))
        self.reset_volume_action.setText(translate('CartLayout', 'Reset volume'))

    @CueLayout.model_adapter.getter
    def model_adapter(self):
        return self._model_adapter

    def add_pages(self):
        pages, accepted = QInputDialog.getInt(
            self,
            translate('CartLayout', 'Add pages'),
            translate('CartLayout', 'Number of Pages:'),
            value=1, min=1, max=10)

        if accepted:
            for _ in range(pages):
                self.add_page()

    def add_page(self):
        page = PageWidget(self.__rows, self.__columns, self)
        page.move_drop_event.connect(self._move_widget)
        page.copy_drop_event.connect(self._copy_widget)

        self.addTab(page, 'Page ' + str(self.count() + 1))
        self.__pages.append(page)

    def select_context_cue(self):
        self.__context_widget.selected = not self.__context_widget.selected

    def select_all(self, cue_class=Cue):
        for widget in self.widgets():
            if isinstance(widget.cue, cue_class):
                widget.selected = True

    def deselect_all(self, cue_class=Cue):
        for widget in self.widgets():
            if isinstance(widget.cue, cue_class):
                widget.selected = False

    def invert_selection(self):
        for widget in self.widgets():
            widget.selected = not widget.selected

    def contextMenuEvent(self, event):
        # For some reason the currentWidget geometry does not include the
        # vertical offset created by the tabBar.
        geometry = self.currentWidget().geometry().translated(0, self.tabBar().geometry().height())
        if geometry.contains(event.pos()):
            self.show_context_menu(event.globalPos())

    def dragEnterEvent(self, event):
        if qApp.keyboardModifiers() == Qt.ControlModifier:
            event.setDropAction(Qt.MoveAction)
            event.accept()
        elif qApp.keyboardModifiers() == Qt.ShiftModifier:
            event.setDropAction(Qt.MoveAction)
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if self.tabBar().contentsRect().contains(event.pos()):
            self.setCurrentIndex(self.tabBar().tabAt(event.pos()))
            event.accept()

    def dropEvent(self, e):
        e.ignore()

    def keyPressEvent(self, event):
        self.key_pressed.emit(event)
        event.ignore()

    def get_context_cue(self):
        if self.__context_widget is not None:
            return self.__context_widget.cue

    def get_selected_cues(self, cue_class=Cue):
        # w -> widget
        cues = []
        for widget in self.widgets():
            if widget.selected:
                cues.append(widget.cue)
        return cues

    def remove_current_page(self):
        if self.__pages:
            confirm = QMessageBox()
            confirm.setIcon(confirm.Question)
            confirm.setWindowTitle(translate('CartLayout', 'Warning'))
            confirm.setText(
                translate('CartLayout', 'Every cue in the page will be lost.'))
            confirm.setInformativeText(
                translate('CartLayout', 'Are you sure to continue?'))
            confirm.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            confirm.setDefaultButton(QMessageBox.No)

            if confirm.exec_() == QMessageBox.Yes:
                self.remove_page(self.currentIndex())

    def remove_page(self, page):
        if len(self.__pages) > page >= 0:
            self._model_adapter.remove_page(page)

            self.removeTab(page)
            self.tabRemoved(page)
            page_widget = self.__pages.pop(page)
            page_widget.move_drop_event.disconnect()
            page_widget.copy_drop_event.disconnect()

            # Rename every successive tab accordingly
            text = translate('CartLayout', 'Page') + ' {}'
            for n in range(page, self.count()):
                self.setTabText(n, text.format(n + 1))

    def widgets(self):
        for page in self.__pages:
            for widget in page.widgets():
                yield widget

    def set_countdown_mode(self, mode):
        self._countdown_mode = mode
        for widget in self.widgets():
            widget.set_countdown_mode(mode)

    def set_accurate(self, enable):
        self._accurate_timing = enable
        for widget in self.widgets():
            widget.set_accurate_timing(enable)

    def set_seek_visible(self, visible):
        self._show_seek = visible
        for widget in self.widgets():
            widget.show_seek_slider(visible)

    def set_dbmeter_visible(self, visible):
        self._show_dbmeter = visible
        for widget in self.widgets():
            widget.show_dbmeters(visible)

    def set_volume_visible(self, visible):
        self._show_volume = visible
        for widget in self.widgets():
            widget.show_volume_slider(visible)

    def to_3d_index(self, index):
        page_size = self.__rows * self.__columns

        page = index // page_size
        row = (index % page_size) // self.__columns
        column = (index % page_size) % self.__columns

        return page, row, column

    def to_1d_index(self, index):
        try:
            page, row, column = index
            page *= self.__rows * self.__columns
            row *= self.__columns
            return page + row + column
        except(TypeError, ValueError):
            return -1

    def finalize(self):
        MainWindow().menuLayout.clear()

        # Disconnect menu-actions signals
        self.edit_action.triggered.disconnect()
        self.remove_action.triggered.disconnect()
        self.select_action.triggered.disconnect()
        self.play_action.triggered.disconnect()
        self.pause_action.triggered.disconnect()
        self.stop_action.triggered.disconnect()
        self.reset_volume_action.disconnect()

        # Remove context-items
        self.cm_registry.remove_item(self.edit_action)
        self.cm_registry.remove_item(self.sep1)
        self.cm_registry.remove_item(self.remove_action)
        self.cm_registry.remove_item(self.select_action)
        self.cm_registry.remove_item(self.sep2)
        self.cm_registry.remove_item(self.play_action)
        self.cm_registry.remove_item(self.pause_action)
        self.cm_registry.remove_item(self.stop_action)
        self.cm_registry.remove_item(self.reset_volume_action)

        # Delete the layout
        self.deleteLater()

    def _move_widget(self, widget, to_row, to_column):
        new_index = self.to_1d_index((self.currentIndex(), to_row, to_column))
        self._model_adapter.move(widget.cue.index, new_index)

    def _copy_widget(self, widget, to_row, to_column):
        new_index = self.to_1d_index((self.currentIndex(), to_row, to_column))
        new_cue = CueFactory.clone_cue(widget.cue)

        self._model_adapter.insert(new_cue, new_index)

    def _play_context_cue(self):
        self.get_context_cue().media.play()

    def _pause_context_cue(self):
        self.get_context_cue().media.pause()

    def _stop_context_cue(self):
        self.get_context_cue().media.stop()

    def _on_context_menu(self, widget, position):
        self.__context_widget = widget
        self.show_cue_context_menu(position)

    def _edit_cue_action(self):
        self.edit_cue(self.get_context_cue())

    def _remove_cue_action(self):
        self._model_adapter.remove(self.get_context_cue())
        self.__context_widget = None

    def _reset_cue_volume(self):
        self.__context_widget.reset_volume()

    def __cue_added(self, cue):
        page, row, column = self.to_3d_index(cue.index)

        widget = CueWidget(cue)
        widget.cue_executed.connect(self.cue_executed.emit)
        widget.context_menu_request.connect(self._on_context_menu)
        widget.edit_request.connect(self.edit_cue)
        widget.set_accurate_timing(self._accurate_timing)
        widget.set_countdown_mode(self._countdown_mode)
        widget.show_dbmeters(self._show_dbmeter)
        widget.show_seek_slider(self._show_seek)
        widget.show_volume_slider(self._show_volume)

        if page >= len(self.__pages):
            self.add_page()

        self.__pages[page].add_widget(widget, row, column)
        self.setCurrentIndex(page)

    def __cue_removed(self, cue):
        if isinstance(cue, MediaCue):
            cue.media.interrupt()
        else:
            cue.stop()

        page, row, column = self.to_3d_index(cue.index)
        widget = self.__pages[page].take_widget(row, column)

        widget.cue_executed.disconnect()
        widget.context_menu_request.disconnect()
        widget.edit_request.disconnect()

        widget.deleteLater()

    def __cue_moved(self, old_index, new_index):
        o_page, o_row, o_column = self.to_3d_index(old_index)
        n_page, n_row, n_column = self.to_3d_index(new_index)

        if o_page == n_page:
            self.__pages[n_page].move_widget(o_row, o_column, n_row, n_column)
        else:
            widget = self.__pages[o_page].take_widget(o_row, o_column)
            self.__pages[n_page].add_widget(widget, n_row, n_column)

    def __model_reset(self):
        self.__context_widget = None
        for page in self.__pages:
            page.reset()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import jack
import logging
from PyQt5.QtCore import QT_TRANSLATE_NOOP

from lisp.backend.media_element import ElementType, MediaType
from lisp.core.has_properties import Property
from lisp.modules.gst_backend.gi_repository import Gst
from lisp.modules.gst_backend.gst_element import GstMediaElement, GstProperty


class JackSink(GstMediaElement):
    ElementType = ElementType.Output
    MediaType = MediaType.Audio
    Name = QT_TRANSLATE_NOOP('MediaElementName', 'JACK Out')

    CLIENT_NAME = 'linux-show-player'
    CONNECT_MODE = 'none'

    _ControlClient = None
    _clients = []

    connections = Property(default=[[] for _ in range(8)])

    def __init__(self, pipeline):
        super().__init__()

        if JackSink._ControlClient is None:
            JackSink._ControlClient = jack.Client('LinuxShowPlayer_Control')

        self.pipeline = pipeline
        self.audio_resample = Gst.ElementFactory.make('audioresample')
        self.jack_sink = Gst.ElementFactory.make('jackaudiosink', 'sink')

        self._client_id = JackSink.__register_client_id()
        self._client_name = JackSink.CLIENT_NAME + '-' + str(self._client_id)
        self.jack_sink.set_property('client-name', self._client_name)
        self.jack_sink.set_property('connect', JackSink.CONNECT_MODE)

        self.pipeline.add(self.audio_resample)
        self.pipeline.add(self.jack_sink)

        self.audio_resample.link(self.jack_sink)

        self.connections = self.default_connections(JackSink._ControlClient)
        self.changed('connections').connect(self.__prepare_connections)

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        self._handler = bus.connect('message', self.__on_message)

    def sink(self):
        return self.audio_resample

    def dispose(self):
        try:
            self.pipeline.get_bus().disconnect(self._handler)
            JackSink._clients.remove(self._client_id)
        finally:
            if not JackSink._clients:
                JackSink._ControlClient.close()
                JackSink._ControlClient = None

    @classmethod
    def default_connections(cls, client):
        # Up to 8 channels
        connections = [[] for _ in range(8)]

        if isinstance(client, jack.Client):
            # Search for default input ports
            input_ports = client.get_ports(name_pattern='^system:',
                                           is_audio=True, is_input=True)
            for n, port in enumerate(input_ports):
                if n < len(connections):
                    connections[n].append(port.name)
                else:
                    break

        return connections

    def __prepare_connections(self, value):
        if (self.pipeline.current_state == Gst.State.PLAYING or
                    self.pipeline.current_state == Gst.State.PAUSED):
            self.__jack_connect()

    @classmethod
    def __register_client_id(cls):
        n = 0
        for n, client in enumerate(cls._clients):
            if n != client:
                break

        if n == len(cls._clients) - 1:
            n += 1

        cls._clients.insert(n, n)
        return n

    def __jack_connect(self):
        out_ports = JackSink._ControlClient.get_ports(
            name_pattern='^' + self._client_name + ':.+', is_audio=True)

        for port in out_ports:
            for conn_port in JackSink._ControlClient.get_all_connections(port):
                try:
                    JackSink._ControlClient.disconnect(port, conn_port)
                except jack.JackError as e:
                    logging.error('GST: JACK-SINK: {}'.format(e))

        for output, in_ports in enumerate(self.connections):
            for input_name in in_ports:
                if output < len(out_ports):
                    try:
                        JackSink._ControlClient.connect(out_ports[output],
                                                        input_name)
                    except jack.JackError as e:
                        logging.error('GST: JACK-SINK: {}'.format(e))
                else:
                    break

    def __on_message(self, bus, message):
        if message.src == self.jack_sink:
            if message.type == Gst.MessageType.STATE_CHANGED:
                change = message.parse_state_changed()

                # The jack ports are available when the the jackaudiosink
                # change from READY to PAUSED state
                if (change[0] == Gst.State.READY and
                        change[1] == Gst.State.PAUSED):
                    self.__jack_connect()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from threading import Lock

from PyQt5.QtCore import QT_TRANSLATE_NOOP

from lisp.core.configuration import config
from lisp.core.decorators import async
from lisp.core.fade_functions import FadeInType, FadeOutType
from lisp.core.fader import Fader
from lisp.core.has_properties import NestedProperties
from lisp.cues.cue import Cue, CueAction, CueState


class MediaCue(Cue):
    Name = QT_TRANSLATE_NOOP('CueName', 'Media Cue')

    _media_ = NestedProperties('media', default={})

    CueActions = (CueAction.Default, CueAction.Start, CueAction.FadeInStart,
                  CueAction.Stop, CueAction.FadeOutStop, CueAction.Pause,
                  CueAction.FadeOut, CueAction.FadeIn, CueAction.FadeOutPause,
                  CueAction.Interrupt, CueAction.FadeOutInterrupt)

    def __init__(self, media, id=None):
        super().__init__(id=id)
        self.default_start_action = CueAction.FadeInStart.value
        self.default_stop_action = CueAction.FadeOutStop.value

        self.media = media
        self.media.changed('duration').connect(self._duration_change)
        self.media.elements_changed.connect(self.__elements_changed)
        self.media.error.connect(self._on_error)
        self.media.eos.connect(self._on_eos)

        self.__in_fadein = False
        self.__in_fadeout = False

        self.__volume = self.media.element('Volume')
        self.__fader = Fader(self.__volume, 'current_volume')
        self.__fade_lock = Lock()

    def __elements_changed(self):
        self.__volume = self.media.element('Volume')
        self.__fader.target = self.__volume

    def __start__(self, fade=False):
        if fade and self._can_fade(self.fadein_duration):
            self.__volume.current_volume = 0

        self.media.play()
        if fade:
            self._on_start_fade()

        return True

    def __stop__(self, fade=False):
        if self.__in_fadeout:
            self.__fader.stop()
        else:
            if self.__in_fadein:
                self.__fader.stop()

            if self._state & CueState.Running and fade:
                self._st_lock.release()
                ended = self._on_stop_fade()
                self._st_lock.acquire()
                if not ended:
                    return False

        self.media.stop()
        return True

    def __pause__(self, fade=False):
        if self.__in_fadeout:
            self.__fader.stop()
        else:
            if self.__in_fadein:
                self.__fader.stop()

            if fade:
                self._st_lock.release()
                ended = self._on_stop_fade()
                self._st_lock.acquire()
                if not ended:
                    return False

        self.media.pause()
        return True

    def __interrupt__(self, fade=False):
        self.__fader.stop()

        if self._state & CueState.Running and fade:
            self._on_stop_fade(interrupt=True)

        self.media.interrupt()

    @async
    def fadein(self, duration, fade_type):
        if not self._st_lock.acquire(timeout=0.1):
            return

        if self._state & CueState.Running:
            self.__fader.stop()

            if self.__volume is not None:
                if duration <= 0:
                    self.__volume.current_volume = self.__volume.volume
                else:
                    self._st_lock.release()
                    self.__fadein(duration, self.__volume.volume, fade_type)
                    return

        self._st_lock.release()

    @async
    def fadeout(self, duration, fade_type):
        if not self._st_lock.acquire(timeout=0.1):
            return

        if self._state & CueState.Running:
            self.__fader.stop()

            if self.__volume is not None:
                if duration <= 0:
                    self.__volume.current_volume = 0
                else:
                    self._st_lock.release()
                    self.__fadeout(duration, 0, fade_type)
                    return

        self._st_lock.release()

    def __fadein(self, duration, to_value, fade_type):
        ended = True
        if self._can_fade(duration):
            with self.__fade_lock:
                self.__in_fadein = True
                self.fadein_start.emit()
                try:
                    self.__fader.prepare()
                    ended = self.__fader.fade(duration, to_value, fade_type)
                finally:
                    self.__in_fadein = False
                    self.fadein_end.emit()

        return ended

    def __fadeout(self, duration, to_value, fade_type):
        ended = True
        if self._can_fade(duration):
            with self.__fade_lock:
                self.__in_fadeout = True
                self.fadeout_start.emit()
                try:
                    self.__fader.prepare()
                    ended = self.__fader.fade(duration, to_value, fade_type)
                finally:
                    self.__in_fadeout = False
                    self.fadeout_end.emit()

        return ended

    def current_time(self):
        return self.media.current_time()

    def _duration_change(self, value):
        self.duration = value

    def _on_eos(self, *args):
        with self._st_lock:
            self.__fader.stop()
            self._ended()

    def _on_error(self, media, message, details):
        with self._st_lock:
            self.__fader.stop()
            self._error(message, details)

    def _can_fade(self, duration):
        return self.__volume is not None and duration > 0

    @async
    def _on_start_fade(self):
        if self.__volume is not None:
            self.__fadein(self.fadein_duration,
                          self.__volume.volume,
                          FadeInType[self.fadein_type])

    def _on_stop_fade(self, interrupt=False):
        if interrupt:
            duration = config['Cue'].getfloat('InterruptFade')
            fade_type = config['Cue'].get('InterruptFadeType')
        else:
            duration = self.fadeout_duration
            fade_type = self.fadeout_type

        return self.__fadeout(duration, 0, FadeOutType[fade_type])



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
from zipfile import ZipFile, BadZipFile

from lisp.core import configuration
from lisp.core.actions_handler import MainActionsHandler
from lisp.cues.cue_actions import UpdateCueAction, UpdateCuesAction

try:
    from os import scandir
except ImportError:
    from scandir import scandir

PRESETS_DIR = os.path.join(configuration.CFG_DIR, 'presets')


def preset_path(name):
    """Return the preset-file path.

    :param name: Name of the preset
    :type name: str
    """
    return os.path.join(PRESETS_DIR, name)


def scan_presets():
    """Iterate over presets.

    Every time this function is called a search in `PRESETS_DIR` is performed.
    """
    for entry in scandir(PRESETS_DIR):
        if entry.is_file():
            yield entry.name


def preset_exists(name):
    """Return True if the preset already exist, False otherwise.

    :param name: Name of the preset
    :type name: str
    :rtype: bool
    """
    return os.path.exists(preset_path(name))


def load_on_cue(preset_name, cue):
    """Load the preset with the given name on cue.

    Use `UpdateCueAction`

    :param preset_name: The preset to be loaded
    :type preset_name: str
    :param cue: The cue on which load the preset
    :type cue: lisp.cue.Cue
    """
    MainActionsHandler.do_action(
        UpdateCueAction(load_preset(preset_name), cue))


def load_on_cues(preset_name, cues):
    """
    Use `UpdateCuesAction`

    :param preset_name: The preset to be loaded
    :type preset_name: str
    :param cues: The cues on which load the preset
    :type cues: typing.Iterable[lisp.cue.Cue]
    """
    MainActionsHandler.do_action(
        UpdateCuesAction(load_preset(preset_name), cues))


def load_preset(name):
    """Load the preset with the given name and return it.

    :param name: The preset name
    :type name: str
    :rtype: dict
    """
    path = preset_path(name)

    with open(path, mode='r') as in_file:
        return json.load(in_file)


def write_preset(name, preset):
    """Write a preset with the given name in `PRESET_DIR`.

    :param name: The preset name
    :type name: str
    :param preset: The content of the preset
    :type preset: dict
    :return: True when no error occurs, False otherwise
    :rtype: bool
    """
    path = preset_path(name)

    with open(path, mode='w') as out_file:
        json.dump(preset, out_file)


def rename_preset(old_name, new_name):
    """Rename an existing preset, if the new name is not already used.

    :param old_name: The preset (old) name
    :param new_name: The new preset name
    :return: True if the preset as been renamed successfully, False otherwise
    :rtype: bool
    """
    os.rename(preset_path(old_name), preset_path(new_name))


def delete_preset(name):
    """Delete a preset form the file-system.

    :param name: The preset to be deleted
    :type name: str
    """
    path = preset_path(name)

    if os.path.exists(path):
        os.remove(path)


class PresetImportError(Exception):
    """
    Raised when an error occur during presets import.
    """


class PresetExportError(Exception):
    """
    Raised when an error occur during presets export.
    """


def export_presets(names, archive):
    """Export presets-files into an archive.

    :param names: The presets to be exported
    :type names: typing.Iterable[Str]
    :param archive: The path of the archive
    :type archive: str
    """
    if names:
        try:
            with ZipFile(archive, mode='w') as archive:
                for name in names:
                    archive.write(preset_path(name), name)
        except(OSError, BadZipFile) as e:
            raise PresetExportError(str(e))


def import_presets(archive, overwrite=True):
    """Import presets for an archive.

    :param archive: The archive path
    :type archive: str
    :param overwrite: Overwrite existing files
    :type overwrite: bool
    """
    try:
        with ZipFile(archive) as archive:
            for member in archive.namelist():
                if not (preset_exists(member) and not overwrite):
                    archive.extract(member, path=PRESETS_DIR)
    except(OSError, BadZipFile) as e:
        raise PresetImportError(str(e))


def import_has_conflicts(archive):
    """Verify if the `archive` files conflicts with the user presets.

    :param archive: path of the archive to analyze
    :type archive: str
    :rtype: bool
    """
    try:
        with ZipFile(archive) as archive:
            for member in archive.namelist():
                if preset_exists(member):
                    return True
    except(OSError, BadZipFile) as e:
        raise PresetImportError(str(e))

    return False


def import_conflicts(archive):
    """Return a list of conflicts between `archive` files and user presets.

    :param archive: path of the archive to analyze
    :type archive: str
    """
    conflicts = []

    try:
        with ZipFile(archive) as archive:
            for member in archive.namelist():
                if preset_exists(member):
                    conflicts.append(member)
    except(OSError, BadZipFile) as e:
        raise PresetImportError(str(e))

    return conflicts



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod

from lisp.core.action import Action


class MementoAction(Action):
    """Actions created by the MementoModel to register model changes."""

    __slots__ = ('_m_model', '_model')

    def __init__(self, m_model, model):
        super().__init__()
        self._m_model = m_model
        self._model = model

    def do(self):
        pass

    def undo(self):
        try:
            self._m_model.lock()
            self.__undo__()
        finally:
            self._m_model.unlock()

    def redo(self):
        try:
            self._m_model.lock()
            self.__redo__()
        finally:
            self._m_model.unlock()

    @abstractmethod
    def __undo__(self):
        pass

    @abstractmethod
    def __redo__(self):
        pass


class AddItemAction(MementoAction):

    __slots__ = '__item'

    def __init__(self, m_model, model, item):
        super().__init__(m_model, model)
        self.__item = item

    def __undo__(self):
        self._model.remove(self.__item)

    def __redo__(self):
        self._model.add(self.__item)


class RemoveItemAction(MementoAction):

    __slots__ = '__item'

    def __init__(self, m_model, model, item):
        super().__init__(m_model, model)
        self.__item = item

    def __undo__(self):
        self._model.add(self.__item)

    def __redo__(self):
        self._model.remove(self.__item)


class MoveItemAction(MementoAction):

    __slots__ = ('__old_index', '__new_index')

    def __init__(self, m_model, model_adapter, old_index, new_index):
        super().__init__(m_model, model_adapter)
        self.__old_index = old_index
        self.__new_index = new_index

    def __undo__(self):
        self._model.move(self.__new_index, self.__old_index)

    def __redo__(self):
        self._model.move(self.__old_index, self.__new_index)



#!/usr/bin/env python3

import os
import re
from setuptools import find_packages, setup

import lisp


def find_files(directory, regex='.*', rel=''):
    """Search for files that match `regex` in `directory`."""
    paths = []
    for path, directories, file_names in os.walk(directory):
        for filename in file_names:
            file_path = os.path.join(path, filename)
            if re.match(regex, file_path):
                paths.append(file_path[len(rel):])
    return paths

# List the directories with icons to be installed
lisp_icons = find_files('lisp/ui/styles/icons', rel='lisp/ui/styles/')

# Setup function
setup(
    name='linux-show-player',
    author=lisp.__author__,
    author_email=lisp.__email__,
    version=lisp.__version__,
    license=lisp.__license__,
    url=lisp.__email__,
    description='Cue player for live shows',
    install_requires=[
        'sortedcontainers',
        'mido',
        'python-rtmidi',
        'JACK-Client',
        'scandir;python_version<"3.5"'
    ],
    packages=find_packages(),
    package_data={
        '': ['i18n/*.qm', '*.qss', '*.cfg'],
        'lisp.ui.styles': lisp_icons,
    },
    scripts=['linux-show-player']
)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QStandardPaths, Qt
from PyQt5.QtWidgets import QGroupBox, QHBoxLayout, QPushButton, QLineEdit, \
    QGridLayout, QCheckBox, QSpinBox, QLabel, QFileDialog, QVBoxLayout

from lisp.modules.gst_backend.elements.uri_input import UriInput
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class UriInputSettings(SettingsPage):
    ELEMENT = UriInput
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.fileGroup = QGroupBox(self)
        self.fileGroup.setLayout(QHBoxLayout())
        self.layout().addWidget(self.fileGroup)

        self.buttonFindFile = QPushButton(self.fileGroup)
        self.fileGroup.layout().addWidget(self.buttonFindFile)

        self.filePath = QLineEdit('file://', self.fileGroup)
        self.fileGroup.layout().addWidget(self.filePath)

        self.bufferingGroup = QGroupBox(self)
        self.bufferingGroup.setLayout(QGridLayout())
        self.layout().addWidget(self.bufferingGroup)

        self.useBuffering = QCheckBox(self.bufferingGroup)
        self.bufferingGroup.layout().addWidget(self.useBuffering, 0, 0, 1, 2)

        self.download = QCheckBox(self.bufferingGroup)
        self.bufferingGroup.layout().addWidget(self.download, 1, 0, 1, 2)

        self.bufferSize = QSpinBox(self.bufferingGroup)
        self.bufferSize.setRange(-1, 2147483647)
        self.bufferSize.setValue(-1)
        self.bufferingGroup.layout().addWidget(self.bufferSize, 2, 0)

        self.bufferSizeLabel = QLabel(self.bufferingGroup)
        self.bufferSizeLabel.setAlignment(Qt.AlignCenter)
        self.bufferingGroup.layout().addWidget(self.bufferSizeLabel, 2, 1)

        self.buttonFindFile.clicked.connect(self.select_file)

        self.retranslateUi()

    def retranslateUi(self):
        self.fileGroup.setTitle(translate('UriInputSettings', 'Source'))
        self.buttonFindFile.setText(translate('UriInputSettings', 'Find File'))
        self.bufferingGroup.setTitle(translate('UriInputSettings', 'Buffering'))
        self.useBuffering.setText(
            translate('UriInputSettings', 'Use Buffering'))
        self.download.setText(translate('UriInputSettings',
                                        'Attempt download on network streams'))
        self.bufferSizeLabel.setText(
            translate('UriInputSettings', 'Buffer size (-1 default value)'))

    def get_settings(self):
        settings = {}

        checkable = self.fileGroup.isCheckable()

        if not (checkable and not self.fileGroup.isChecked()):
            settings['uri'] = self.filePath.text()
        if not (checkable and not self.bufferingGroup.isChecked()):
            settings['use_buffering'] = self.useBuffering.isChecked()
            settings['download'] = self.download.isChecked()
            settings['buffer_size'] = self.bufferSize.value()

        return settings

    def load_settings(self, settings):
        self.filePath.setText(settings.get('uri', ''))

    def enable_check(self, enable):
        self.fileGroup.setCheckable(enable)
        self.fileGroup.setChecked(False)

    def select_file(self):
        path = QStandardPaths.writableLocation(QStandardPaths.MusicLocation)
        file, ok = QFileDialog.getOpenFileName(self,
                                               translate('UriInputSettings',
                                                         'Choose file'),
                                               path,
                                               translate('UriInputSettings',
                                                         'All files') + ' (*)')

        if ok:
            self.filePath.setText('file://' + file)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt, QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QGroupBox, QHBoxLayout, QLabel, QComboBox,\
    QVBoxLayout, QDoubleSpinBox, QTabWidget, QWidget

from lisp.cues.cue import CueNextAction, CueAction
from lisp.ui.settings.settings_page import CueSettingsPage
from lisp.ui.ui_utils import translate
from lisp.ui.widgets import FadeComboBox
from lisp.ui.widgets.fade_edit import FadeEdit


class CueGeneralSettings(CueSettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Cue')

    def __init__(self, cue_class, **kwargs):
        super().__init__(cue_class, **kwargs)
        self.setLayout(QVBoxLayout())

        self.tabWidget = QTabWidget(self)
        self.layout().addWidget(self.tabWidget)

        # TAB 1 (Behaviours)
        self.tab_1 = QWidget(self.tabWidget)
        self.tab_1.setLayout(QVBoxLayout())
        self.tabWidget.addTab(self.tab_1, '1')

        # Start-Action
        self.startActionGroup = QGroupBox(self.tab_1)
        self.startActionGroup.setLayout(QHBoxLayout())
        self.tab_1.layout().addWidget(self.startActionGroup)

        self.startActionCombo = QComboBox(self.startActionGroup)
        for action in [CueAction.Start, CueAction.FadeInStart]:
            if action in cue_class.CueActions:
                self.startActionCombo.addItem(
                    translate('CueAction', action.name), action.value)
        self.startActionCombo.setEnabled(self.startActionCombo.count() > 1)
        self.startActionGroup.layout().addWidget(self.startActionCombo)

        self.startActionLabel = QLabel(self.startActionGroup)
        self.startActionLabel.setAlignment(Qt.AlignCenter)
        self.startActionGroup.layout().addWidget(self.startActionLabel)

        # Stop-Action
        self.stopActionGroup = QGroupBox(self.tab_1)
        self.stopActionGroup.setLayout(QHBoxLayout())
        self.tab_1.layout().addWidget(self.stopActionGroup)

        self.stopActionCombo = QComboBox(self.stopActionGroup)
        for action in [CueAction.Stop, CueAction.Pause,
                       CueAction.FadeOutStop, CueAction.FadeOutPause]:
            if action in cue_class.CueActions:
                self.stopActionCombo.addItem(
                    translate('CueAction', action.name), action.value)
        self.stopActionCombo.setEnabled(self.stopActionCombo.count() > 1)
        self.stopActionGroup.layout().addWidget(self.stopActionCombo)

        self.stopActionLabel = QLabel(self.stopActionGroup)
        self.stopActionLabel.setAlignment(Qt.AlignCenter)
        self.stopActionGroup.layout().addWidget(self.stopActionLabel)

        self.tab_1.layout().addSpacing(150)
        self.tab_1.setEnabled(self.stopActionCombo.isEnabled() and
                              self.startActionCombo.isEnabled())

        # TAB 2 (Pre/Post Wait)
        self.tab_2 = QWidget(self.tabWidget)
        self.tab_2.setLayout(QVBoxLayout())
        self.tabWidget.addTab(self.tab_2, '2')

        # Pre wait
        self.preWaitGroup = QGroupBox(self.tab_2)
        self.preWaitGroup.setLayout(QHBoxLayout())
        self.tab_2.layout().addWidget(self.preWaitGroup)

        self.preWaitSpin = QDoubleSpinBox(self.preWaitGroup)
        self.preWaitSpin.setMaximum(3600 * 24)
        self.preWaitGroup.layout().addWidget(self.preWaitSpin)

        self.preWaitLabel = QLabel(self.preWaitGroup)
        self.preWaitLabel.setAlignment(Qt.AlignCenter)
        self.preWaitGroup.layout().addWidget(self.preWaitLabel)

        # Post wait
        self.postWaitGroup = QGroupBox(self.tab_2)
        self.postWaitGroup.setLayout(QHBoxLayout())
        self.tab_2.layout().addWidget(self.postWaitGroup)

        self.postWaitSpin = QDoubleSpinBox(self.postWaitGroup)
        self.postWaitSpin.setMaximum(3600 * 24)
        self.postWaitGroup.layout().addWidget(self.postWaitSpin)

        self.postWaitLabel = QLabel(self.postWaitGroup)
        self.postWaitLabel.setAlignment(Qt.AlignCenter)
        self.postWaitGroup.layout().addWidget(self.postWaitLabel)

        # Next action
        self.nextActionGroup = QGroupBox(self.tab_2)
        self.nextActionGroup.setLayout(QHBoxLayout())
        self.tab_2.layout().addWidget(self.nextActionGroup)

        self.nextActionCombo = QComboBox(self.nextActionGroup)
        for action in CueNextAction:
            self.nextActionCombo.addItem(
                translate('CueNextAction', action.name), action.value)
        self.nextActionGroup.layout().addWidget(self.nextActionCombo)

        # TAB 3 (Fade In/Out)
        self.tab_3 = QWidget(self.tabWidget)
        self.tab_3.setLayout(QVBoxLayout())
        self.tabWidget.addTab(self.tab_3, '3')

        # FadeIn
        self.fadeInGroup = QGroupBox(self.tab_3)
        self.fadeInGroup.setEnabled(
            CueAction.FadeInStart in cue_class.CueActions
        )
        self.fadeInGroup.setLayout(QHBoxLayout())
        self.tab_3.layout().addWidget(self.fadeInGroup)

        self.fadeInEdit = FadeEdit(self.fadeInGroup,
                                   mode=FadeComboBox.Mode.FadeIn)
        self.fadeInGroup.layout().addWidget(self.fadeInEdit)

        # FadeOut
        self.fadeOutGroup = QGroupBox(self.tab_3)
        self.fadeOutGroup.setEnabled(
            CueAction.FadeOutPause in cue_class.CueActions or
            CueAction.FadeOutStop in cue_class.CueActions
        )
        self.fadeOutGroup.setLayout(QHBoxLayout())
        self.tab_3.layout().addWidget(self.fadeOutGroup)

        self.fadeOutEdit = FadeEdit(self.fadeOutGroup,
                                    mode=FadeComboBox.Mode.FadeOut)
        self.fadeOutGroup.layout().addWidget(self.fadeOutEdit)

        self.retranslateUi()

    def retranslateUi(self):
        # Tabs
        self.tabWidget.setTabText(0, translate('CueSettings', 'Behaviours'))
        self.tabWidget.setTabText(1, translate('CueSettings', 'Pre/Post Wait'))
        self.tabWidget.setTabText(2, translate('CueSettings', 'Fade In/Out'))

        # Start-Action
        self.startActionGroup.setTitle(
            translate('CueSettings', 'Start action'))
        self.startActionLabel.setText(
            translate('CueSettings', 'Default action to start the cue'))
        # Stop-Action
        self.stopActionGroup.setTitle(
            translate('CueSettings', 'Stop action'))
        self.stopActionLabel.setText(
            translate('CueSettings', 'Default action to stop the cue'))

        # PreWait
        self.preWaitGroup.setTitle(translate('CueSettings', 'Pre wait'))
        self.preWaitLabel.setText(
            translate('CueSettings', 'Wait before cue execution'))
        # PostWait
        self.postWaitGroup.setTitle(translate('CueSettings', 'Post wait'))
        self.postWaitLabel.setText(
            translate('CueSettings', 'Wait after cue execution'))
        # NextAction
        self.nextActionGroup.setTitle(translate('CueSettings', 'Next action'))

        # FadeIn/Out
        self.fadeInGroup.setTitle(translate('FadeSettings', 'Fade In'))
        self.fadeOutGroup.setTitle(translate('FadeSettings', 'Fade Out'))

    def load_settings(self, settings):
        self.startActionCombo.setCurrentText(
            translate('CueAction', settings.get('default_start_action', '')))
        self.stopActionCombo.setCurrentText(
            translate('CueAction', settings.get('default_stop_action', '')))

        self.preWaitSpin.setValue(settings.get('pre_wait', 0))
        self.postWaitSpin.setValue(settings.get('post_wait', 0))
        self.nextActionCombo.setCurrentText(
            translate('CueNextAction', settings.get('next_action', '')))

        self.fadeInEdit.setFadeType(settings.get('fadein_type', ''))
        self.fadeInEdit.setDuration(settings.get('fadein_duration', 0))
        self.fadeOutEdit.setFadeType(settings.get('fadeout_type', ''))
        self.fadeOutEdit.setDuration(settings.get('fadeout_duration', 0))

    def enable_check(self, enable):
        self.startActionGroup.setCheckable(enable)
        self.startActionGroup.setChecked(False)
        self.stopActionGroup.setCheckable(enable)
        self.stopActionGroup.setChecked(False)

        self.preWaitGroup.setCheckable(enable)
        self.preWaitGroup.setChecked(False)
        self.postWaitGroup.setCheckable(enable)
        self.postWaitGroup.setChecked(False)
        self.nextActionGroup.setCheckable(enable)
        self.nextActionGroup.setChecked(False)

        self.fadeInGroup.setCheckable(enable)
        self.fadeInGroup.setChecked(False)
        self.fadeOutGroup.setCheckable(enable)
        self.fadeOutGroup.setChecked(False)

    def get_settings(self):
        conf = {}
        checkable = self.preWaitGroup.isCheckable()

        if not (checkable and not self.startActionGroup.isChecked()):
            if self.startActionCombo.isEnabled():
                conf['default_start_action'] = self.startActionCombo.currentData()
        if not (checkable and not self.stopActionGroup.isChecked()):
            if self.stopActionCombo.isEnabled():
                conf['default_stop_action'] = self.stopActionCombo.currentData()

        if not (checkable and not self.preWaitGroup.isChecked()):
            conf['pre_wait'] = self.preWaitSpin.value()
        if not (checkable and not self.postWaitGroup.isChecked()):
            conf['post_wait'] = self.postWaitSpin.value()
        if not (checkable and not self.nextActionGroup.isChecked()):
            conf['next_action'] = self.nextActionCombo.currentData()

        if not (checkable and not self.fadeInGroup.isChecked()):
            conf['fadein_type'] = self.fadeInEdit.fadeType()
            conf['fadein_duration'] = self.fadeInEdit.duration()
        if not (checkable and not self.fadeInGroup.isChecked()):
            conf['fadeout_type'] = self.fadeOutEdit.fadeType()
            conf['fadeout_duration'] = self.fadeOutEdit.duration()

        return conf



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod

from lisp.core.model import Model, ModelException


class ProxyModel(Model):
    """Proxy that wrap a more generic model to extend its functionality.

    The add, remove, __iter__, __len__ and __contains__ default implementations
    use the the model ones.

    .. note:
        The base model cannot be changed.
        Any ProxyModel could provide it's own methods/signals.
    """

    def __init__(self, model):
        super().__init__()

        if not isinstance(model, Model):
            raise TypeError('ProxyModel model must be a Model object, not {0}'
                            .format(model.__class__.__name__))

        self.__model = model
        self.__model.item_added.connect(self._item_added)
        self.__model.item_removed.connect(self._item_removed)
        self.__model.model_reset.connect(self._model_reset)

    def add(self, item):
        self.__model.add(item)

    def remove(self, item):
        self.__model.remove(item)

    def reset(self):
        self.__model.reset()

    @property
    def model(self):
        return self.__model

    @abstractmethod
    def _model_reset(self):
        pass

    @abstractmethod
    def _item_added(self, item):
        pass

    @abstractmethod
    def _item_removed(self, item):
        pass

    def __iter__(self):
        return self.__model.__iter__()

    def __len__(self):
        return len(self.__model)

    def __contains__(self, item):
        return item in self.__model


class ReadOnlyProxyModel(ProxyModel):
    def add(self, item):
        raise ModelException('cannot add items into a read-only model')

    def remove(self, item):
        raise ModelException('cannot remove items from a read-only model')

    def reset(self):
        raise ModelException('cannot reset read-only model')



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from lisp.application import Application
from lisp.core.has_properties import Property
from lisp.core.plugin import Plugin
from lisp.cues.cue import Cue, CueAction
from lisp.plugins.controller import protocols
from lisp.plugins.controller.controller_settings import ControllerSettings
from lisp.ui.settings.cue_settings import CueSettingsRegistry


class Controller(Plugin):

    Name = 'Controller'

    def __init__(self):
        super().__init__()
        self.__map = {}
        self.__actions_map = {}
        self.__protocols = {}

        # Register a new Cue property to store settings
        Cue.register_property('controller', Property(default={}))

        # Listen cue_model changes
        Application().cue_model.item_added.connect(self.__cue_added)
        Application().cue_model.item_removed.connect(self.__cue_removed)

        # Register settings-page
        CueSettingsRegistry().add_item(ControllerSettings)
        # Load available protocols
        self.__load_protocols()

    def init(self):
        for protocol in self.__protocols.values():
            protocol.init()

    def reset(self):
        self.__map.clear()
        self.__actions_map.clear()

        for protocol in self.__protocols.values():
            protocol.reset()

    def cue_changed(self, cue, property_name, value):
        if property_name == 'controller':
            self.delete_from_map(cue)

            for protocol in self.__protocols:
                for key, action in value.get(protocol, []):
                    if key not in self.__map:
                        self.__map[key] = set()

                    self.__map[key].add(cue)
                    self.__actions_map[(key, cue)] = CueAction(action)

    def delete_from_map(self, cue):
        for key in self.__map:
            self.__map[key].discard(cue)
            self.__actions_map.pop((key, cue), None)

    def perform_action(self, key):
        for cue in self.__map.get(key, []):
            cue.execute(self.__actions_map[(key, cue)])

    def __cue_added(self, cue):
        cue.property_changed.connect(self.cue_changed)
        self.cue_changed(cue, 'controller', cue.controller)

    def __cue_removed(self, cue):
        cue.property_changed.disconnect(self.cue_changed)
        self.delete_from_map(cue)

    def __load_protocols(self):
        protocols.load()

        for protocol_class in protocols.Protocols:
            try:
                protocol = protocol_class()
                protocol.protocol_event.connect(self.perform_action)

                self.__protocols[protocol_class.__name__.lower()] = protocol
            except Exception as e:
                import logging
                import traceback

                logging.error('CONTROLLER: cannot setup protocol "{}"'.format(
                    protocol_class.__name__))
                logging.debug(traceback.format_exc())



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.


# In the current state set/get_backend are quite useless
__backend = None


def set_backend(backend):
    global __backend
    __backend = backend


def get_backend():
    """
    :rtype: lisp.backends.base.backend.Backend
    """
    return __backend




# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QPushButton


class QMuteButton(QPushButton):
    def __init__(self, *args):
        super().__init__(*args)

        # Use the button-check behaviors as mute-unmute
        self.setCheckable(True)
        self.setChecked(False)
        # Set the icon
        self.onToggle()

        # More explicit names
        self.isMute = self.isChecked
        self.setMute = self.setChecked

        self.toggled.connect(self.onToggle)

    def onToggle(self):
        if self.isChecked():
            self.setIcon(QIcon.fromTheme('audio-volume-muted'))
        else:
            self.setIcon(QIcon.fromTheme('audio-volume-high'))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QWidget


class SettingsPage(QWidget):
    Name = 'Page'

    def enable_check(self, enabled):
        """Enable option check"""

    def get_settings(self):
        """Return the current settings."""
        return {}

    def load_settings(self, settings):
        """Load the settings."""


class CueSettingsPage(SettingsPage):
    Name = 'Cue page'

    def __init__(self, cue_class, **kwargs):
        super().__init__(**kwargs)

        self._cue_class = cue_class



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod

from lisp.core.singleton import ABCSingleton


class Backend(metaclass=ABCSingleton):
    """Common interface that any backend must provide.

    An __init__() method can be defined if the backend require initialization.
    """

    @abstractmethod
    def uri_duration(self, uri):
        """Return the file duration in milliseconds.

        :param uri: The URI of the file
        :type uri: str

        :rtype: int
        """

    @abstractmethod
    def uri_tags(self, uri):
        """Return a dictionary containing the file metadata/tags.

        :param uri: The URI of the file
        :type uri: str

        :rtype: dict
        """

    @abstractmethod
    def supported_extensions(self):
        """Return file extensions supported by the backend.

        Extensions will be categorized in 'audio' and 'video', optionally the
        backend can create others categories.
        e.g. {'audio': ['wav', 'mp3', ...], 'video': ['mp4', 'mov', ...]}

        :rtype: dict
        """



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.


def json_deep_search(iterable, field):
    """Takes a JSON like data-structure and search for the occurrences
    of the given field/key.
    """
    fields_found = []

    if isinstance(iterable, dict):
        for key, value in iterable.items():
            if key == field:
                fields_found.append(value)

            elif isinstance(value, (dict, list)):
                fields_found.extend(json_deep_search(value, field))

    elif isinstance(iterable, list):
        for item in iterable:
            fields_found.extend(json_deep_search(item, field))

    return fields_found


def json_deep_replace(iterable, field, replace_function):
    """Takes a JSON like data-structure and replace the value for all the
    occurrences of the given field/key with the given replace-function.
    """
    if isinstance(iterable, dict):
        for key, value in iterable.items():
            if key == field:
                iterable[key] = replace_function(value)

            elif isinstance(value, (dict, list)):
                json_deep_replace(value, field, replace_function)

    elif isinstance(iterable, list):
        for item in iterable:
            json_deep_replace(item, field, replace_function)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import os

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QComboBox, QPushButton, QFrame, QTextBrowser, QFileDialog, QGridLayout

from lisp import layouts
from lisp.ui.ui_utils import translate


class LayoutSelect(QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.filepath = ''

        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle(translate('LayoutSelect', 'Layout selection'))
        self.setMaximumSize(675, 300)
        self.setMinimumSize(675, 300)
        self.resize(675, 300)

        self.setLayout(QGridLayout(self))
        self.layout().setContentsMargins(5, 5, 5, 5)

        self.layoutCombo = QComboBox(self)
        self.layoutCombo.currentIndexChanged.connect(self.show_description)
        self.layout().addWidget(self.layoutCombo, 0, 0)

        self.confirmButton = QPushButton(self)
        self.confirmButton.setText(translate('LayoutSelect', 'Select layout'))
        self.layout().addWidget(self.confirmButton, 0, 1)

        self.fileButton = QPushButton(self)
        self.fileButton.setText(translate('LayoutSelect', 'Open file'))
        self.layout().addWidget(self.fileButton, 0, 2)

        self.layout().setColumnStretch(0, 3)
        self.layout().setColumnStretch(1, 2)
        self.layout().setColumnStretch(2, 1)

        line = QFrame(self)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        self.layout().addWidget(line, 1, 0, 1, 3)

        self.description = QTextBrowser(self)
        self.layout().addWidget(self.description, 2, 0, 1, 3)

        for layout_class in layouts.get_layouts():
            self.layoutCombo.addItem(layout_class.NAME, layout_class)

        if self.layoutCombo.count() == 0:
            raise Exception('No layout installed!')

        self.confirmButton.clicked.connect(self.accept)
        self.fileButton.clicked.connect(self.open_file)

    def selected(self):
        return self.layoutCombo.currentData()

    def show_description(self, index):
        layout = self.layoutCombo.currentData()

        details = '<ul>'
        for detail in layout.DETAILS:
            details += '<li>' + translate('LayoutDetails', detail)
        details += '</ul>'

        self.description.setHtml(
            '<center><h2>' + layout.NAME + '</h2>'
            '<i><h4>' + translate('LayoutDescription', layout.DESCRIPTION) +
            '</h4></i></center>' + details)

    def open_file(self):
        path, _ = QFileDialog.getOpenFileName(self, filter='*.lsp',
                                              directory=os.getenv('HOME'))
        self.filepath = path
        self.accept()
